import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
import scipy.special as sp
from patsy import dmatrices, dmatrix, demo_data

from .. import families as fam
from .. import tsm as tsm
from .. import data_check as dc

from .gas_core_recursions import gas_reg_recursion

class GASReg(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** GENERALIZED AUTOREGRESSIVE SCORE (GAS) REGRESSION MODELS ****

    Parameters
    ----------

    formula : string
        patsy string describing the regression

    data : pd.DataFrame or np.array
        Field to specify the data that will be used

    family : GAS family object
        Which distribution to use, e.g. GASNormal()

    """

    def __init__(self, formula, data, family):

        # Initialize TSM object     
        super(GASReg,self).__init__('GASReg')

        # Latent Variables
        self.max_lag = 0
        self._z_hide = 0 # Whether to cutoff variance latent variables from results
        self.supported_methods = ["MLE","PML","Laplace","M-H","BBVI"]
        self.default_method = "MLE"
        self.multivariate_model = False
        self.skewness = False

        # Format the data
        self.is_pandas = True # This is compulsory for this model type
        self.data_original = data
        self.formula = formula
        self.y, self.X = dmatrices(formula, data)
        self.y_name = self.y.design_info.describe()
        self.X_names = self.X.design_info.describe().split(" + ")
        self.y = self.y.astype(np.float) 
        self.X = self.X.astype(np.float) 
        self.z_no = self.X.shape[1]
        self.data_name = self.y_name
        self.y = np.array([self.y]).ravel()
        self.data = self.y
        self.X = np.array([self.X])[0]
        self.index = data.index
        self.initial_values = np.zeros(self.z_no)

        self.data_length = self.data.shape[0]
        self._create_model_matrices()
        self._create_latent_variables()

        self.family = family
        
        self.model_name2, self.link, self.scale, self.shape, self.skewness, self.mean_transform, self.cythonized = self.family.setup()
    
        # Identify whether model has cythonized backend - then choose update type
        if self.cythonized is True:
            self._model = self._cythonized_model 
            self._mb_model = self._cythonized_mb_model
            self.recursion = self.family.gradientreg_recursion()
        else:
            self._model = self._uncythonized_model
            self._mb_model = self._uncythonized_mb_model

        self.model_name = self.model_name2 + " GAS Regression"

        # Build any remaining latent variables that are specific to the family chosen
        for no, i in enumerate(self.family.build_latent_variables()):
            self.latent_variables.add_z(i[0],i[1],i[2])
            self.latent_variables.z_list[no+self.z_no].start = i[3]

        self.family_z_no = len(self.family.build_latent_variables())
        self.z_no += len(self.family.build_latent_variables())

    def _create_model_matrices(self):
        """ Creates model matrices/vectors

        Returns
        ----------
        None (changes model attributes)
        """

        self.model_Y = self.data
        self.model_scores = np.zeros((self.X.shape[1], self.model_Y.shape[0]+1))

    def _create_latent_variables(self):
        """ Creates model latent variables

        Returns
        ----------
        None (changes model attributes)
        """

        for parm in range(self.z_no):
            self.latent_variables.add_z('Scale ' + self.X_names[parm], fam.Flat(transform='exp'), fam.Normal(0, 3))
            self.latent_variables.z_list[parm].start = -5.0
        self.z_no = len(self.latent_variables.z_list)

    def _get_scale_and_shape(self,parm):
        """ Obtains appropriate model scale and shape latent variables

        Parameters
        ----------
        parm : np.array
            Transformed latent variable vector

        Returns
        ----------
        None (changes model attributes)
        """

        if self.scale is True:
            if self.shape is True:
                model_shape = parm[-1]  
                model_scale = parm[-2]
            else:
                model_shape = 0
                model_scale = parm[-1]
        else:
            model_scale = 0
            model_shape = 0 
        if self.skewness is True:
            model_skewness = parm[-3]
        else:
            model_skewness = 0
        return model_scale, model_shape, model_skewness

    def _get_scale_and_shape_sim(self, transformed_lvs):
        """ Obtains model scale, shape, skewness latent variables for
        a 2d array of simulations.

        Parameters
        ----------
        transformed_lvs : np.array
            Transformed latent variable vector (2d - with draws of each variable)

        Returns
        ----------
        - Tuple of np.arrays (each being scale, shape and skewness draws)
        """

        if self.scale is True:
            if self.shape is True:
                model_shape = self.latent_variables.z_list[-1].prior.transform(transformed_lvs[-1, :]) 
                model_scale = self.latent_variables.z_list[-2].prior.transform(transformed_lvs[-2, :])
            else:
                model_shape = np.zeros(transformed_lvs.shape[1])
                model_scale = self.latent_variables.z_list[-1].prior.transform(transformed_lvs[-1, :])
        else:
            model_scale = np.zeros(transformed_lvs.shape[1])
            model_shape = np.zeros(transformed_lvs.shape[1])

        if self.skewness is True:
            model_skewness = self.latent_variables.z_list[-3].prior.transform(transformed_lvs[-3, :])
        else:
            model_skewness = np.zeros(transformed_lvs.shape[1])

        return model_scale, model_shape, model_skewness

    def _cythonized_model(self, beta):
        """ Creates the structure of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        theta : np.array
            Contains the predicted values for the time series

        Y : np.array
            Contains the length-adjusted time series (accounting for lags)

        scores : np.array
            Contains the scores for the time series
        """

        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        coefficients = np.zeros((self.X.shape[1],self.model_Y.shape[0]+1))
        coefficients[:,0] = self.initial_values
        theta = np.zeros(self.model_Y.shape[0]+1)
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)

        # Loop over time series
        theta, self.model_scores, coefficients = self.recursion(parm, theta, self.X, coefficients, self.model_scores, self.model_Y, self.model_Y.shape[0], model_scale, model_shape, model_skewness)

        return np.array(theta[:-1]), self.model_Y, self.model_scores, coefficients

    def _cythonized_mb_model(self, beta, mini_batch):
        """ Creates the structure of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        mini_batch : int
            Size of each mini batch of data

        Returns
        ----------
        theta : np.array
            Contains the predicted values for the time series

        Y : np.array
            Contains the length-adjusted time series (accounting for lags)

        scores : np.array
            Contains the scores for the time series
        """

        rand_int =  np.random.randint(low=0, high=self.data_length-mini_batch-self.max_lag+1)
        sample = np.arange(start=rand_int, stop=rand_int+mini_batch)

        data = self.data[sample]
        X = self.X[sample, :]
        Y = data[self.max_lag:]

        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        coefficients = np.zeros((X.shape[1], Y.shape[0]+1))
        coefficients[:,0] = self.initial_values
        theta = np.zeros(Y.shape[0]+1)
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)

        # Loop over time series
        theta, self.model_scores, coefficients = self.recursion(parm, theta, X, coefficients, self.model_scores, Y, Y.shape[0], model_scale, model_shape, model_skewness)

        return np.array(theta[:-1]), Y, self.model_scores, coefficients

    def _uncythonized_model(self, beta):
        """ Creates the structure of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        theta : np.array
            Contains the predicted values for the time series

        Y : np.array
            Contains the length-adjusted time series (accounting for lags)

        scores : np.array
            Contains the scores for the time series
        """

        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        coefficients = np.zeros((self.X.shape[1],self.model_Y.shape[0]+1))
        coefficients[:,0] = self.initial_values
        theta = np.zeros(self.model_Y.shape[0]+1)
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)

        # Loop over time series
        theta, self.model_scores, coefficients = gas_reg_recursion(parm, theta, self.X, coefficients, self.model_scores, self.model_Y, self.model_Y.shape[0], 
            self.family.reg_score_function, self.link, model_scale, model_shape, model_skewness, self.max_lag)

        return theta[:-1], self.model_Y, self.model_scores, coefficients

    def _uncythonized_model(self, beta, mini_batch):
        """ Creates the structure of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        mini_batch : int
            Size of each mini batch of data

        Returns
        ----------
        theta : np.array
            Contains the predicted values for the time series

        Y : np.array
            Contains the length-adjusted time series (accounting for lags)

        scores : np.array
            Contains the scores for the time series
        """

        rand_int =  np.random.randint(low=0, high=self.data_length-mini_batch-self.max_lag+1)
        sample = np.arange(start=rand_int, stop=rand_int+mini_batch)

        data = self.data[sample]
        X = self.X[sample, :]
        Y = data[self.max_lag:]

        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        coefficients = np.zeros((X.shape[1], Y.shape[0]+1))
        coefficients[:,0] = self.initial_values
        theta = np.zeros(Y.shape[0]+1)
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)

        # Loop over time series
        theta, self.model_scores, coefficients = gas_reg_recursion(parm, theta, X, coefficients, self.model_scores, Y, Y.shape[0], 
            self.family.reg_score_function, self.link, model_scale, model_shape, model_skewness, self.max_lag)

        return theta[:-1], Y, self.model_scores, coefficients

    def _preoptimize_model(self, initials, method):
        """ Preoptimizes the model by estimating a static model, then a quick search of good dynamic parameters

        Parameters
        ----------
        initials : np.array
            A vector of inital values

        method : str
            One of 'MLE' or 'PML' (the optimization options)

        Returns
        ----------
        Y_exp : np.array
            Vector of past values and predictions 
        """
        
        # Random search for good starting values
        start_values = []
        start_values.append(np.ones(len(self.X_names))*-2.0)
        start_values.append(np.ones(len(self.X_names))*-3.0)
        start_values.append(np.ones(len(self.X_names))*-4.0)
        start_values.append(np.ones(len(self.X_names))*-5.0)

        best_start = self.latent_variables.get_z_starting_values()
        best_lik = self.neg_loglik(self.latent_variables.get_z_starting_values())
        proposal_start = best_start.copy()

        for start in start_values:
            proposal_start[:len(self.X_names)] = start
            proposal_likelihood = self.neg_loglik(proposal_start)
            if proposal_likelihood < best_lik:
                best_lik = proposal_likelihood
                best_start = proposal_start.copy()

        return best_start

    def neg_loglik(self, beta):
        """ Returns the negative loglikelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        """
        theta, Y, scores,_ = self._model(beta)
        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)
        return self.family.neg_loglikelihood(Y,self.link(theta),model_scale,model_shape,model_skewness)

    def mb_neg_loglik(self, beta, mini_batch):
        """ Returns the negative loglikelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        mini_batch : int
            Size of each mini batch of data
        """
        theta, Y, scores,_ = self._mb_model(beta, mini_batch)
        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)
        return self.family.neg_loglikelihood(Y,self.link(theta),model_scale,model_shape,model_skewness)

    def plot_fit(self, **kwargs):
        """ Plots the fit of the model

        Notes
        ----------
        Intervals are bootstrapped as follows: take the filtered values from the
        algorithm (thetas). Use these thetas to generate a pseudo data stream from
        the measurement density. Use the GAS algorithm and estimated latent variables to
        filter the pseudo data. Repeat this N times. 

        Returns
        ----------
        None (plots data and the fit)
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:

            date_index = self.index.copy()
            mu, Y, scores, coefficients = self._model(self.latent_variables.get_z_values())

            if self.model_name2 == "Exponential":
                values_to_plot = 1.0/self.link(mu)
            elif self.model_name2 == "Skewt":
                t_params = self.transform_z()
                model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_params)
                m1 = (np.sqrt(model_shape)*sp.gamma((model_shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(model_shape/2.0))
                additional_loc = (model_skewness - (1.0/model_skewness))*model_scale*m1
                values_to_plot = mu + additional_loc
            else:
                values_to_plot = self.link(mu)

            plt.figure(figsize=figsize) 
            
            plt.subplot(len(self.X_names)+1, 1, 1)
            plt.title(self.y_name + " Filtered")
            plt.plot(date_index,Y,label='Data')
            plt.plot(date_index,values_to_plot,label='GAS Filter',c='black')
            plt.legend(loc=2)

            for coef in range(0,len(self.X_names)):
                plt.subplot(len(self.X_names)+1, 1, 2+coef)
                plt.title("Beta " + self.X_names[coef]) 
                plt.plot(date_index,coefficients[coef,0:-1],label='Coefficient')
                plt.legend(loc=2)               

            plt.show()          
    
    def plot_predict(self, h=5, past_values=20, intervals=True, oos_data=None, **kwargs):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show prediction intervals for the forecast?

        oos_data : pd.DataFrame
            Data for the variables to be used out of sample (ys can be NaNs)

        Returns
        ----------
        - Plot of the forecast
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            # Sort/manipulate the out-of-sample data
            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            X_pred = X_oos[:h]

            date_index = self.shift_dates(h)

            if self.latent_variables.estimation_method in ['M-H']:
                
                sim_vector = np.zeros([15000,h])

                for n in range(0, 15000):
                    t_z = self.draw_latent_variables(nsims=1).T[0]
                    _, Y, _, coefficients = self._model(t_z)
                    coefficients_star = coefficients.T[-1]
                    theta_pred = np.dot(np.array([coefficients_star]), X_pred.T)[0]
                    t_z = np.array([self.latent_variables.z_list[k].prior.transform(t_z[k]) for k in range(t_z.shape[0])])
                    model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_z)
                    sim_vector[n,:] = self.family.draw_variable(self.link(theta_pred), model_scale, model_shape, model_skewness, theta_pred.shape[0])
                mean_values = np.append(Y, self.link(np.array([np.mean(i) for i in sim_vector.T])))
            else:

                # Retrieve data, dates and (transformed) latent variables
                _, Y, _, coefficients = self._model(self.latent_variables.get_z_values()) 
                coefficients_star = coefficients.T[-1] 
                theta_pred = np.dot(np.array([coefficients_star]), X_pred.T)[0]  
                t_z = self.transform_z()
                sim_vector = np.zeros([15000,h])
                mean_values = np.append(Y, self.link(theta_pred))
                model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_z)

                if self.model_name2 == "Skewt":
                    m1 = (np.sqrt(model_shape)*sp.gamma((model_shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(model_shape/2.0))
                    mean_values += (model_skewness - (1.0/model_skewness))*model_scale*m1 

                for n in range(0,15000):
                    sim_vector[n,:] = self.family.draw_variable(self.link(theta_pred),model_scale,model_shape,model_skewness,theta_pred.shape[0])

            sim_vector = sim_vector.T
            error_bars = []
            for pre in range(5,100,5):
                error_bars.append(np.insert([np.percentile(i,pre) for i in sim_vector], 0, mean_values[-h-1]))
            forecasted_values = mean_values[-h-1:]
            plot_values = mean_values[-h-past_values:]
            plot_index = date_index[-h-past_values:]

            plt.figure(figsize=figsize)
            if intervals == True:
                alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                for count in range(9):
                    plt.fill_between(date_index[-h-1:], error_bars[count], error_bars[-count],
                        alpha=alpha[count])     
            plt.plot(plot_index,plot_values)
            plt.title("Forecast for " + self.data_name)
            plt.xlabel("Time")
            plt.ylabel(self.data_name)
            plt.show()

    def plot_predict_is(self, h=5, fit_once=True, fit_method='MLE', **kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        fit_once : boolean
            (default: True) Fits only once before the in-sample prediction; if False, fits after every new datapoint

        fit_method : string
            Which method to fit the model with

        Returns
        ----------
        - Plot of the forecast against data 
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        predictions = self.predict_is(h=h, fit_method=fit_method, fit_once=fit_once)
        data = self.data[-h:]
        plt.plot(predictions.index,data,label='Data')
        plt.plot(predictions.index,predictions,label='Predictions',c='black')
        plt.title(self.y_name)
        plt.legend(loc=2)   
        plt.show()          

    def predict(self, h=5, oos_data=None, intervals=False, **kwargs):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        oos_data : pd.DataFrame
            Data for the variables to be used out of sample (ys can be NaNs)

        intervals : boolean (default: False)
            Whether to return prediction intervals

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     
        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            # Sort/manipulate the out-of-sample data
            _, X_oos = dmatrices(self.formula, oos_data)
            X_oos = np.array([X_oos])[0]
            X_pred = X_oos[:h]

            date_index = self.shift_dates(h)

            if self.latent_variables.estimation_method in ['M-H']:
                
                sim_vector = np.zeros([15000,h])

                for n in range(0, 15000):
                    t_z = self.draw_latent_variables(nsims=1).T[0]
                    _, Y, _, coefficients = self._model(t_z)
                    coefficients_star = coefficients.T[-1]
                    theta_pred = np.dot(np.array([coefficients_star]), X_pred.T)[0]
                    t_z = np.array([self.latent_variables.z_list[k].prior.transform(t_z[k]) for k in range(t_z.shape[0])])
                    model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_z)
                    sim_vector[n,:] = self.family.draw_variable(self.link(theta_pred), model_scale, model_shape, model_skewness, theta_pred.shape[0])

                sim_vector = sim_vector.T

                forecasted_values = np.array([np.mean(i) for i in sim_vector])
                prediction_01 = np.array([np.percentile(i, 1) for i in sim_vector])
                prediction_05 = np.array([np.percentile(i, 5) for i in sim_vector])
                prediction_95 = np.array([np.percentile(i, 95) for i in sim_vector])
                prediction_99 = np.array([np.percentile(i, 99) for i in sim_vector])

            else:

                # Retrieve data, dates and (transformed) latent variables
                _, Y, _, coefficients = self._model(self.latent_variables.get_z_values()) 
                coefficients_star = coefficients.T[-1] 
                theta_pred = np.dot(np.array([coefficients_star]), X_pred.T)[0]  
                t_z = self.transform_z()
                mean_values = np.append(Y, self.link(theta_pred))
                model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_z)

                if self.model_name2 == "Skewt":
                    m1 = (np.sqrt(model_shape)*sp.gamma((model_shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(model_shape/2.0))
                    forecasted_values = mean_values[-h:] + (model_skewness - (1.0/model_skewness))*model_scale*m1 
                else:
                    forecasted_values = mean_values[-h:] 

            if intervals is False:
                result = pd.DataFrame(forecasted_values)
                result.rename(columns={0:self.data_name}, inplace=True)
            else:
                # Get mean prediction and simulations (for errors)
                if self.latent_variables.estimation_method not in ['M-H']:
                    sim_values = np.zeros([15000,h])

                    if intervals is True:
                        for n in range(0,15000):
                            sim_values[n,:] = self.family.draw_variable(self.link(theta_pred),model_scale,model_shape,model_skewness,theta_pred.shape[0])

                    sim_values = sim_values.T

                    prediction_01 = np.array([np.percentile(i, 1) for i in sim_values])
                    prediction_05 = np.array([np.percentile(i, 5) for i in sim_values])
                    prediction_95 = np.array([np.percentile(i, 95) for i in sim_values])
                    prediction_99 = np.array([np.percentile(i, 99) for i in sim_values])

                result = pd.DataFrame([forecasted_values, prediction_01, prediction_05, 
                    prediction_95, prediction_99]).T
                result.rename(columns={0:self.data_name, 1: "1% Prediction Interval", 
                    2: "5% Prediction Interval", 3: "95% Prediction Interval", 4: "99% Prediction Interval"}, 
                    inplace=True)
 
            result.index = date_index[-h:]

            return result

    def predict_is(self, h=5, fit_once=True, fit_method='MLE', intervals=False):
        """ Makes dynamic in-sample predictions with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        fit_once : boolean
            (default: True) Fits only once before the in-sample prediction; if False, fits after every new datapoint

        fit_method : string
            Which method to fit the model with

        intervals : boolean (default: False)
            Whether to return prediction intervals

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        predictions = []

        for t in range(0,h):
            data1 = self.data_original.iloc[:-(h+t),:]
            data2 = self.data_original.iloc[-h+t:,:]
            x = GASReg(formula=self.formula, data=self.data_original[:(-h+t)], family=self.family)

            if fit_once is False:
                x.fit(method=fit_method, printer=False)
            if t == 0:
                if fit_once is True:
                    x.fit(method=fit_method, printer=False)
                    saved_lvs = x.latent_variables
                predictions = x.predict(h=1, oos_data=data2, intervals=intervals)
            else:
                if fit_once is True:
                    x.latent_variables = saved_lvs
                predictions = pd.concat([predictions,x.predict(h=1, oos_data=data2, intervals=intervals)])

        predictions.rename(columns={0:self.y_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions

    def sample(self, nsims=1000):
        """ Samples from the posterior predictive distribution

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws from the posterior predictive distribution

        Returns
        ----------
        - np.ndarray of draws from the data
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            lv_draws = self.draw_latent_variables(nsims=nsims)
            mus = [self._model(lv_draws[:,i])[0] for i in range(nsims)]
            model_scale, model_shape, model_skewness = self._get_scale_and_shape_sim(lv_draws)
            data_draws = np.array([self.family.draw_variable(self.link(mus[i]), 
                np.repeat(model_scale[i], mus[i].shape[0]), np.repeat(model_shape[i], mus[i].shape[0]), 
                np.repeat(model_skewness[i], mus[i].shape[0]), mus[i].shape[0]) for i in range(nsims)])
            return data_draws

    def plot_sample(self, nsims=10, plot_data=True, **kwargs):
        """
        Plots draws from the posterior predictive density against the data

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws from the posterior predictive distribution

        plot_data boolean
            Whether to plot the data or not
        """

        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            import matplotlib.pyplot as plt
            import seaborn as sns

            figsize = kwargs.get('figsize',(10,7))
            plt.figure(figsize=figsize)
            date_index = self.index.copy()
            _, Y, _, _ = self._model(self.latent_variables.get_z_values()) 
            draws = self.sample(nsims).T
            plt.plot(date_index, draws, label='Posterior Draws', alpha=1.0)
            if plot_data is True:
                plt.plot(date_index, Y, label='Data', c='black', alpha=0.5, linestyle='', marker='s')
            plt.title(self.data_name)
            plt.show()    

    def ppc(self, nsims=1000, T=np.mean):
        """ Computes posterior predictive p-value

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws for the PPC

        T : function
            A discrepancy measure - e.g. np.mean, np.std, np.max

        Returns
        ----------
        - float (posterior predictive p-value)
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            lv_draws = self.draw_latent_variables(nsims=nsims)
            mus = [self._model(lv_draws[:,i])[0] for i in range(nsims)]
            model_scale, model_shape, model_skewness = self._get_scale_and_shape_sim(lv_draws)
            data_draws = np.array([self.family.draw_variable(self.link(mus[i]), 
                np.repeat(model_scale[i], mus[i].shape[0]), np.repeat(model_shape[i], mus[i].shape[0]), 
                np.repeat(model_skewness[i], mus[i].shape[0]), mus[i].shape[0]) for i in range(nsims)])
            T_sims = T(self.sample(nsims=nsims), axis=1)
            T_actual = T(self.data)
            return len(T_sims[T_sims>T_actual])/nsims

    def plot_ppc(self, nsims=1000, T=np.mean, **kwargs):
        """ Plots histogram of the discrepancy from draws of the posterior

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws for the PPC

        T : function
            A discrepancy measure - e.g. np.mean, np.std, np.max
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            import matplotlib.pyplot as plt
            import seaborn as sns

            figsize = kwargs.get('figsize',(10,7))

            lv_draws = self.draw_latent_variables(nsims=nsims)
            mus = [self._model(lv_draws[:,i])[0] for i in range(nsims)]
            model_scale, model_shape, model_skewness = self._get_scale_and_shape_sim(lv_draws)
            data_draws = np.array([self.family.draw_variable(self.link(mus[i]), 
                np.repeat(model_scale[i], mus[i].shape[0]), np.repeat(model_shape[i], mus[i].shape[0]), 
                np.repeat(model_skewness[i], mus[i].shape[0]), mus[i].shape[0]) for i in range(nsims)])
            T_sim = T(self.sample(nsims=nsims), axis=1)
            T_actual = T(self.data)

            if T == np.mean:
                description = " of the mean"
            elif T == np.max:
                description = " of the maximum"
            elif T == np.min:
                description = " of the minimum"
            elif T == np.median:
                description = " of the median"
            else:
                description = ""

            plt.figure(figsize=figsize)
            ax = plt.subplot()
            ax.axvline(T_actual)
            sns.distplot(T_sim, kde=False, ax=ax)
            ax.set(title='Posterior predictive' + description, xlabel='T(x)', ylabel='Frequency');
            plt.show()


import numpy as np
import pandas as pd

from .covariances import acf
from .output import TablePrinter
from .inference import *
from .families import Normal, t, TruncatedNormal, Cauchy, Skewt, InverseGamma, Flat, InverseWishart, Laplace

class LatentVariables(object):
    """ Latent Variables Class

    Holds latent variable objects and contains method for latent variable manipulation. Latent variables are 
    referred to as z as shorthand. This is convention in much of the literature.
    """

    def __init__(self,model_name):
        self.model_name = model_name
        self.z_list = []
        self.z_indices = {}
        self.estimated = False
        self.estimation_method = None

    def __str__(self):
        z_row = []
        z_names = self.get_z_names()
        priors = self.get_z_priors()
        prior_names, prior_z_names = self.get_z_priors_names()
        vardist_names = self.get_z_approx_dist_names()
        transforms = self.get_z_transforms_names()

        fmt = [
            ('Index','z_index',8),      
            ('Latent Variable','z_name',25),
            ('Prior','z_prior',15),
            ('Prior Hyperparameters','z_hyper',25),
            ('V.I. Dist','z_vardist',10),
            ('Transform','z_transform',10)
            ]       

        for z in range(len(self.z_list)):
            z_row.append({'z_index': z, 'z_name': z_names[z],
                'z_prior': prior_names[z], 'z_hyper': prior_z_names[z],
                'z_vardist': vardist_names[z], 'z_transform': transforms[z]})
        return( TablePrinter(fmt, ul='=')(z_row) )

    def add_z(self, name, prior, q, index=True):
        """ Adds latent variable

        Parameters
        ----------
        name : str
            Name of the latent variable

        prior : Prior object
            Which prior distribution? E.g. Normal(0,1)

        q : Distribution object
            Which distribution to use for variational approximation

        index : boolean
            Whether to index the variable in the z_indices dictionary

        Returns
        ----------
        None (changes priors in LatentVariables object)
        """

        self.z_list.append(LatentVariable(name,len(self.z_list),prior,q))
        if index is True:
            self.z_indices[name] = {'start': len(self.z_list)-1, 'end': len(self.z_list)-1}

    def create(self, name, dim, prior, q):
        """ Creates multiple latent variables

        Parameters
        ----------
        name : str
            Name of the latent variable

        dim : list
            Dimension of the latent variable arrays

        prior : Prior object
            Which prior distribution? E.g. Normal(0,1)

        q : Distribution object
            Which distribution to use for variational approximation

        Returns
        ----------
        None (changes priors in LatentVariables object)
        """   

        def rec(dim, prev=[]):
           if len(dim) > 0:
               return [rec(dim[1:], prev + [i]) for i in range(dim[0])]
           else:
               return "(" + ",".join([str(j) for j in prev]) + ")" 

        indices = rec(dim)

        for f_dim in range(1, len(dim)):
            indices = sum(indices, [])

        if self.z_list is None:
            starting_index = 0
        else:
            starting_index = len(self.z_list)

        self.z_indices[name] = {'start': starting_index, 'end': starting_index+len(indices)-1, 'dim': len(dim)}

        for index in indices:
            self.add_z(name + " " + index, prior, q, index=False)

    def adjust_prior(self,index,prior):
        """ Adjusts priors for the latent variables

        Parameters
        ----------
        index : int or list[int]
            Which latent variable index/indices to be altered

        prior : Prior object
            Which prior distribution? E.g. Normal(0,1)

        Returns
        ----------
        None (changes priors in Parameters object)
        """

        if isinstance(index, list):
            for item in index:
                if item < 0 or item > (len(self.z_list)-1) or not isinstance(item, int):
                    raise ValueError("Oops - the latent variable index " + str(item) + " you have entered is invalid!")
                else:
                    self.z_list[item].prior = prior 
                    if hasattr(self.z_list[item].prior, 'mu0'):
                        self.z_list[item].start = self.z_list[item].prior.mu0    
                    elif hasattr(self.z_list[item].prior, 'loc0'):
                        self.z_list[item].start = self.z_list[item].prior.loc0
        else:
            if index < 0 or index > (len(self.z_list)-1) or not isinstance(index, int):
                raise ValueError("Oops - the latent variable index " + str(index) + " you have entered is invalid!")
            else:
                self.z_list[index].prior = prior
                if hasattr(self.z_list[index].prior, 'mu0'):
                    self.z_list[index].start = self.z_list[index].prior.mu0  
                elif hasattr(self.z_list[index].prior, 'loc0'):
                    self.z_list[index].start = self.z_list[index].prior.loc0  

    def get_z_names(self):
        names = []
        for z in self.z_list:
            names.append(z.name)
        return names

    def get_z_priors(self):
        priors = []
        for z in self.z_list:
            priors.append(z.prior)
        return priors

    def get_z_priors_names(self):
        priors = self.get_z_priors()
        prior_names = []
        prior_z_names = []
        for prior in priors:
            if isinstance(prior, Normal):
                prior_names.append('Normal')
                prior_z_names.append('mu0: ' + str(np.round(prior.mu0,4)) + ', sigma0: ' + str(np.round(prior.sigma0,4)))
            elif isinstance(prior, Laplace):
                prior_names.append('Laplace')
                prior_z_names.append('loc0: ' + str(np.round(prior.loc0,4)) + ', scale0: ' + str(np.round(prior.scale0,4)))
            elif isinstance(prior, InverseGamma):
                prior_names.append('Inverse Gamma')
                prior_z_names.append('alpha: ' + str(np.round(prior.alpha,4)) + ', beta: ' + str(np.round(prior.beta,4)))
            elif isinstance(prior, Flat):
                prior_names.append('Flat')
                prior_z_names.append('n/a (non-informative)')
            elif isinstance(prior, InverseWishart):
                prior_names.append('InverseWishart')
                prior_z_names.append('v: ' + str(np.round(prior.v,4)) + ' and scale matrix')
            elif isinstance(prior, t):
                prior_names.append('t')
                prior_z_names.append('loc0: ' + str(np.round(prior.loc0,4)) + ', scale0: ' + str(np.round(prior.scale0,4)) 
                    + ', df0: ' + str(np.round(prior.df0,4))) 
            elif isinstance(prior, Skewt):
                prior_names.append('Skewt')
                prior_z_names.append('loc0: ' + str(np.round(prior.loc0,4)) + ', scale0: ' + str(np.round(prior.scale0,4)) 
                    + ', df0: ' + str(np.round(prior.df0,4)) + ', gamma0: ' + str(np.round(prior.gamma0,4))) 
            elif isinstance(prior, Cauchy):
                prior_names.append('Cauchy')
                prior_z_names.append('loc0: ' + str(np.round(prior.loc0,4)) + ', scale0: ' + str(np.round(prior.scale0,4)))
            elif isinstance(prior, TruncatedNormal):
                prior_names.append('TruncatedNormal')
                prior_z_names.append('mu0: ' + str(np.round(prior.mu0,4)) + ', sigma0: ' + str(np.round(prior.sigma0,4)))                 
            else:
                raise ValueError("Prior distribution not detected!")
        return prior_names, prior_z_names

    def get_z_transforms(self):
        transforms = []
        for z in self.z_list:
            transforms.append(z.prior.transform)
        return transforms

    def get_z_transforms_names(self):
        transforms = []
        for z in self.z_list:
            transforms.append(z.prior.transform_name)
        return transforms

    def get_z_starting_values(self,transformed=False):
        transforms = self.get_z_transforms()

        values = np.zeros(len(self.z_list))
        for i in range(len(self.z_list)):
            if transformed is True:
                values[i] = transforms[i](self.z_list[i].start)
            else:
                values[i] = self.z_list[i].start
        return values

    def get_z_values(self, transformed=False):
        transforms = self.get_z_transforms()

        if self.estimated is True:
            values = np.zeros(len(self.z_list))
            for i in range(len(self.z_list)):
                if transformed is True:
                    values[i] = transforms[i](self.z_list[i].value)
                else:
                    values[i] = self.z_list[i].value
            return values
        else:
            return ValueError("No latent variables have been estimated yet")

    def get_z_approx_dist(self):
        dists = []
        for z in self.z_list:
            dists.append(z.q)
        return dists

    def get_z_approx_dist_names(self):
        approx_dists = self.get_z_approx_dist()
        q_list = []

        for approx in approx_dists:
            if isinstance(approx, Normal):
                q_list.append('Normal')
            else:
                raise Exception("Approximate distribution not detected!")
        return q_list

    def set_z_values(self,values,method,std=None,sample=None):
        if len(values) != len(self.z_list):
            raise ValueError("Length of your list is not equal to number of latent variables")
        for no in range(len(self.z_list)):
            self.z_list[no].method = method
            self.z_list[no].value = values[no]
            if std is not None:
                self.z_list[no].std = std[no]
            if sample is not None:
                self.z_list[no].sample = sample[no]
            self.estimated = True

    def set_z_starting_values(self,values):
        if values.shape[0] != len(self.z_list):
            raise ValueError("Length of your array is not equal to number of latent variables")
        for no in range(len(self.z_list)):
            self.z_list[no].start = values[no]

    def plot_z(self,indices=None,figsize=(15,5),loc=1):
        import matplotlib.pyplot as plt
        import matplotlib.mlab as mlab
        import seaborn as sns

        plt.figure(figsize=figsize) 
        for z in range(1,len(self.z_list)+1):
            if indices is not None and z-1 not in indices:
                continue
            else:
                if hasattr(self.z_list[z-1], 'sample'):
                    sns.distplot(self.z_list[z-1].prior.transform(self.z_list[z-1].sample), rug=False, hist=False,label=self.z_list[z-1].method + ' estimate of ' + self.z_list[z-1].name)

                elif hasattr(self.z_list[z-1], 'value') and hasattr(self.z_list[z-1], 'std'): 

                    if self.z_list[z-1].prior.transform_name is None:
                        x = np.linspace(self.z_list[z-1].value-self.z_list[z-1].std*3.5,self.z_list[z-1].value+self.z_list[z-1].std*3.5,100)
                        plt.plot(x,mlab.normpdf(x,self.z_list[z-1].value,self.z_list[z-1].std),label=self.z_list[z-1].method + ' estimate of ' + self.z_list[z-1].name)
                    else:
                        sims = self.z_list[z-1].prior.transform(np.random.normal(self.z_list[z-1].value,self.z_list[z-1].std,100000))
                        sns.distplot(sims, rug=False, hist=False,label=self.z_list[z-1].method + ' estimate of ' + self.z_list[z-1].name)


                else:
                    raise ValueError("No information on latent variable to plot!")        

        plt.xlabel('Value')
        plt.ylabel('Frequency')
        plt.title('Latent Variable Plot')
        plt.legend(loc=1)
        plt.show()

    def trace_plot(self,figsize=(15,15)):
        import matplotlib.pyplot as plt
        import matplotlib.mlab as mlab
        import seaborn as sns

        if hasattr(self.z_list[0], 'sample'):
            fig = plt.figure(figsize=figsize)
            
            palette = [(0.2980392156862745, 0.4470588235294118, 0.6901960784313725), 
            (0.3333333333333333, 0.6588235294117647, 0.40784313725490196), 
            (0.7686274509803922, 0.3058823529411765, 0.3215686274509804), 
            (0.5058823529411764, 0.4470588235294118, 0.6980392156862745), 
            (0.8, 0.7254901960784313, 0.4549019607843137), 
            (0.39215686274509803, 0.7098039215686275, 0.803921568627451)] * len(self.z_list)
            
            for j in range(len(self.z_list)):
                chain = self.z_list[j].sample
                for k in range(4):
                    iteration = j*4 + k + 1
                    ax = fig.add_subplot(len(self.z_list),4,iteration)
                    if iteration in range(1,len(self.z_list)*4 + 1,4):
                        a = sns.distplot(self.z_list[j].prior.transform(chain), rug=False, hist=False,color=palette[j])
                        a.set_ylabel(self.z_list[j].name)
                        if iteration == 1:
                            a.set_title('Density Estimate')
                    elif iteration in range(2,len(self.z_list)*4 + 1,4):
                        a = plt.plot(self.z_list[j].prior.transform(chain),color=palette[j])
                        if iteration == 2:
                            plt.title('Trace Plot')
                    elif iteration in range(3,len(self.z_list)*4 + 1,4): 
                        plt.plot(np.cumsum(self.z_list[j].prior.transform(chain))/np.array(range(1,len(chain)+1)),color=palette[j])
                        if iteration == 3:
                            plt.title('Cumulative Average')                 
                    elif iteration in range(4,len(self.z_list)*4 + 1,4):
                        plt.bar(range(1,10),[acf(chain,lag) for lag in range(1,10)],color=palette[j])
                        if iteration == 4:
                            plt.title('ACF Plot')                       
            sns.plt.show()  
        else:
            raise ValueError("No samples to plot!")


class LatentVariable(object):
    """ LatentVariable Class

    Parameters
    ----------
    name : str
        Name of the latent variable

    index : int
        Index of the latent variable

    prior : Prior object
        The prior for the latent variable, e.g. Normal(0,1)

    q : fam.[dist] object
        The variational distribution for the latent variable, e.g. Normal(0,1)
    """

    def __init__(self,name,index,prior,q):
        self.name = name
        self.index = 0
        self.prior = prior
        self.transform = self.prior.transform
        self.start = 0.0
        self.q = q

    def plot_z(self,figsize=(15,5)):
        import matplotlib.pyplot as plt
        import seaborn as sns

        if hasattr(self, 'sample'):
            sns.distplot(self.prior.transform(self.sample), rug=False, hist=False,label=self.method + ' estimate of ' + self.name)

        elif hasattr(self, 'value') and hasattr(self, 'std'):
            x = np.linspace(self.value-self.std*3.5,self.value+self.std*3.5,100)
            plt.figure(figsize=figsize)
            if self.prior.transform_name is None:
                plt.plot(x,mlab.normpdf(x,self.value,self.std),label=self.method + ' estimate of ' + self.name)
            else:
                sims = self.prior.transform(np.random.normal(self.value,self.std,100000))
                sns.distplot(sims, rug=False, hist=False,label=self.method + ' estimate of ' + self.name)
            plt.xlabel('Value')
            plt.legend()
            plt.show()

        else:
            raise ValueError("No information on latent variable to plot!")


import pandas as pd
import numpy as np

class Aggregate():
    """ Aggregation Algorithm

    Parameters
    ----------
    learning_rate : float
        Learning rate for the Aggregation algorithm

    loss_type : string
        'absolute' or 'squared'

    match_window : int
        (default: 10) how many of the observations at the end of a model time 
        series to assess whether the data being used in each model is the same
    """

    def __init__(self, learning_rate=1.0, loss_type='absolute', match_window=10):
        self.learning_rate = learning_rate
        self.data = []
        self.model_list = []
        self.model_names = []
        self.match_window = match_window
        self.model_predictions_is = []
        self.model_predictions = []

        if loss_type == 'absolute':
            self.loss_type = self.absolute_loss
            self.loss_name = 'Absolute Loss'
        elif loss_type == 'squared':
            self.loss_type = self.squared_loss
            self.loss_name = 'Squared Loss'
        else:
            raise ValueError('Unidentified loss type entered!')

        self.supported_models = ['EGARCH', 'EGARCHM', 'EGARCHMReg', 'GARCH', 'LMEGARCH', 'LMSEGARCH', 'SEGARCH', 'SEGARCHM', 'GAS', 'ARIMA', 'ARIMAX', 'GASLLEV', 'GASLLT', 'GASReg', 'GASX', 'GPNARX', 'LLEV', 'LLT', 'DynReg']

    @staticmethod
    def absolute_loss(data, predictions):
        """ Calculates absolute loss 
        
        Parameters
        ----------
        data : np.ndarray
            Univariate data

        predictions : np.ndarray
            Univariate predictions

        Returns
        ----------
        - np.ndarray of the absolute loss
        """
        return np.abs(data-predictions)     

    @staticmethod
    def squared_loss(data, predictions):
        """ Calculates squared loss
        
        Parameters
        ----------
        data : np.ndarray
            Univariate data

        predictions : np.ndarray
            Univariate predictions

        Returns
        ----------
        - np.ndarray of the squared loss
        """
        return np.square(data-predictions)

    def add_model(self, model):
        """ Adds a PyFlux model to the aggregating algorithm
        
        Parameters
        ----------
        model : pf.[MODEL]
            A PyFlux univariate model

        Returns
        ----------
        - Void (changes self.model_list)
        """

        if model.model_type not in self.supported_models:
            raise ValueError('Model type not supported for Aggregate! Apologies')

        if not self.model_list:
            self.model_list.append(model)
            if model.model_type in ['EGARCH', 'EGARCHM', 'EGARCHMReg', 'GARCH', 'LMEGARCH', 'LMSEGARCH', 'SEGARCH', 'SEGARCHM']:
                self.data = np.abs(model.data)
            else:
                self.data = model.data
            self.index = model.index
        else:
            if model.model_type in ['EGARCH', 'EGARCHM', 'EGARCHMReg', 'GARCH', 'LMEGARCH', 'LMSEGARCH', 'SEGARCH', 'SEGARCHM']:
                if np.isclose(np.abs(np.abs(model.data[-self.match_window:])-self.data[-self.match_window:]).sum(),0.0) or model.model_type=='GPNARX':
                    self.model_list.append(model)
                else:
                    raise ValueError('Data entered is deemed different based on %s last values!' % (s))
            else:
                if np.isclose(np.abs(model.data[-self.match_window:]-self.data[-self.match_window:]).sum(),0.0) or model.model_type=='GPNARX':
                    self.model_list.append(model)
                else:
                    raise ValueError('Data entered is deemed different based on %s last values!' % (s))

        self.model_names = [i.model_name for i in self.model_list]

    def _model_predict(self, h, recalculate=False, fit_once=True):
        """ Outputs ensemble model predictions for out-of-sample data
        
        Parameters
        ----------
        h : int
            How many steps at the end of the series to run the ensemble on

        recalculate: boolean
            Whether to recalculate the predictions or not

        fit_once : boolean
            Whether to fit the model once at the beginning, or with every iteration
        
        Returns
        ----------
        - pd.DataFrame of the model predictions, index of dates
        """

        if len(self.model_predictions) == 0 or h != self.h or recalculate is True:

            for no, model in enumerate(self.model_list):
                if no == 0:
                    model.fit()
                    result = model.predict(h)
                    self.predict_index = result.index
                    result.columns = [model.model_name]
                else:
                    model.fit()
                    new_frame = model.predict(h)
                    new_frame.columns = [model.model_name]
                    result = pd.concat([result,new_frame], axis=1)

            self.model_predictions = result
            self.h = h
            return result, self.predict_index

        else:
            return self.model_predictions, self.predict_index

    def _model_predict_is(self, h, recalculate=False, fit_once=True):
        """ Outputs ensemble model predictions for the end-of-period data
        
        Parameters
        ----------
        h : int
            How many steps at the end of the series to run the ensemble on

        recalculate: boolean
            Whether to recalculate the predictions or not

        fit_once : boolean
            Whether to fit the model once at the beginning, or with every iteration
        
        Returns
        ----------
        - pd.DataFrame of the model predictions, index of dates
        """

        if len(self.model_predictions_is) == 0 or h != self.h or recalculate is True:

            for no, model in enumerate(self.model_list):
                if no == 0:
                    result = model.predict_is(h, fit_once=fit_once)
                    result.columns = [model.model_name]
                else:
                    new_frame = model.predict_is(h, fit_once=fit_once)
                    new_frame.columns = [model.model_name]
                    result = pd.concat([result,new_frame], axis=1)

            self.model_predictions_is = result
            self.h = h
            return result

        else:
            return self.model_predictions_is

    def _construct_losses(self, data, predictions, ensemble_prediction):
        """ Construct losses for the ensemble and each constitute model
        
        Parameters
        ----------
        data: np.ndarray
            The univariate time series

        predictions : np.ndarray
            The predictions of each constitute model

        ensemble_prediction : np.ndarray
            The prediction of the ensemble model

        Returns
        ----------
        - np.ndarray of the losses for each model
        """

        losses = []
        losses.append(self.loss_type(data, ensemble_prediction).sum()/data.shape[0])
        for model in range(len(self.model_list)):
            losses.append(self.loss_type(data, predictions[:,model]).sum()/data.shape[0])
        return losses

    def tune_learning_rate(self, h, parameter_list=None):
        """ Naive tuning of the the learning rate on the in-sample data
        
        Parameters
        ----------
        h : int
            How many steps to run Aggregate on

        parameter_list: list
            List of parameters to search for a good learning rate over

        Returns
        ----------
        - Void (changes self.learning_rate)
        """

        if parameter_list is None:
            parameter_list = [0.001, 0.01, 0.1, 1.0, 10.0, 100.0, 1000.0, 10000.0,100000.0]
        
        for parameter in parameter_list:
            self.learning_rate = parameter
            _, losses, _ = self.run(h, recalculate=False)
            loss = losses[0]
            if parameter == parameter_list[0]:
                best_rate = parameter
                best_loss = loss
            else:
                if loss < best_loss:
                    best_loss = loss
                    best_rate = parameter

        self.learning_rate = best_rate

    def run(self, h, recalculate=False):
        """ Run the aggregating algorithm 
        
        Parameters
        ----------
        h : int
            How many steps to run the aggregating algorithm on 

        recalculate: boolean
            Whether to recalculate the predictions or not

        Returns
        ----------
        - np.ndarray of normalized weights, np.ndarray of losses for each model
        """

        data = self.data[-h:] 
        predictions = self._model_predict_is(h, recalculate=recalculate).values
        weights = np.zeros((h, len(self.model_list)))
        normalized_weights = np.zeros((h, len(self.model_list)))
        ensemble_prediction = np.zeros(h)
        
        for t in range(h):
            if t == 0:
                weights[t,:] = 100000
                ensemble_prediction[t] = np.dot(weights[t,:]/weights[t,:].sum(), predictions[t,:])
                weights[t,:] = weights[t,:]*np.exp(-self.learning_rate*self.loss_type(data[t], predictions[t,:]))
                normalized_weights[t,:] = weights[t,:]/weights[t,:].sum()
            else:
                ensemble_prediction[t] = np.dot(weights[t-1,:]/weights[t-1,:].sum(), predictions[t,:])
                weights[t,:] = weights[t-1,:]*np.exp(-self.learning_rate*self.loss_type(data[t], predictions[t,:]))
                normalized_weights[t,:] = weights[t,:]/weights[t,:].sum()

        return normalized_weights, self._construct_losses(data, predictions, ensemble_prediction), ensemble_prediction

    def plot_weights(self, h, **kwargs):
        """ Plot the weights from the aggregating algorithm
        
        Parameters
        ----------
        h : int
            How many steps to run the aggregating algorithm on 

        Returns
        ----------
        - A plot of the weights for each model constituent over time
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))

        weights, _, _ = self.run(h=h)
        plt.figure(figsize=figsize)
        plt.plot(self.index[-h:],weights)
        plt.legend(self.model_names)
        plt.show()

    def predict(self, h, h_train=40):
        """ Run out-of-sample predicitons for Aggregate algorithm
        (This only works for non-exogenous variable models currently)
        
        Parameters
        ----------
        h : int
            How many out-of-sample steps to run the aggregating algorithm on 

        h_train : int
            How many in-sample steps to warm-up the ensemble weights on

        Returns
        ----------
        - pd.DataFrame of Aggregate out-of-sample predictions
        """

        predictions, index = self._model_predict(h)
        normalized_weights = self.run(h=h_train)[0][-1, :]
        ensemble_prediction = np.zeros(h)
        
        for t in range(h):
            ensemble_prediction[t] = np.dot(normalized_weights, predictions.values[t,:])

        result = pd.DataFrame(ensemble_prediction)
        result.index = index

        return result

    def predict_is(self, h):
        """ Outputs predictions for the Aggregate algorithm on the in-sample data

        Parameters
        ----------
        h : int
            How many steps to run the aggregating algorithm on 

        Returns
        ----------
        - pd.DataFrame of ensemble predictions
        """
        result = pd.DataFrame([self.run(h=h)[2]]).T  
        result.index = self.index[-h:]
        return result 

    def summary(self, h):
        """ 
        Summarize the results for each model for h steps of the algorithm

       Parameters
        ----------
        h : int
            How many steps to run the aggregating algorithm on 

        Returns
        ----------
        - pd.DataFrame of losses for each model       
        """
        _, losses, _ = self.run(h=h)
        df = pd.DataFrame(losses)
        df.index = ['Ensemble'] + self.model_names
        df.columns = [self.loss_name]
        return df



from .cauchy import Cauchy
from .exponential import Exponential
from .flat import Flat
from .inverse_gamma import InverseGamma
from .inverse_wishart import InverseWishart
from .laplace import Laplace
from .normal import Normal
from .poisson import Poisson
from .skewt import Skewt
from .t import t
from .truncated_normal import TruncatedNormal



import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
from patsy import dmatrices, dmatrix, demo_data

from .. import families as fam
from .. import output as op
from .. import tests as tst
from .. import tsm as tsm
from .. import data_check as dc

from .nn_architecture import neural_network_tanh, neural_network_tanh_mb

class NNARX(tsm.TSM):
    """ Inherits time series methods from TSM parent class.

    **** NEURAL NETWORK AUTOREGRESSIVE (NNAR) MODELS ****

    Parameters
    ----------
    data : pd.DataFrame or np.ndarray
        Field to specify the univariate time series data that will be used.

    ar : int
        Field to specify how many AR lags the model will have.

    units : int
        How many units for the neural network

    layers : int
        How many layers for the neural networ

    activation : function
        E.g. np.tanh, np.array (linear)

    family: family object
        E.g. pf.Normal()

    """

    def __init__(self, data, formula, ar, units, layers, family=fam.Normal()):

        # Initialize TSM object
        super(NNARX, self).__init__('NNARX')

        # Latent Variable information
        self.ar = ar
        self.units = units
        self.layers = layers
        self.activation = np.tanh
        self.model_name = "NNARX(" + str(self.ar) + ")"
        self.z_no = self.ar + 2
        self.max_lag = self.ar
        self._z_hide = 0 # Whether to cutoff latent variables from results table
        self.supported_methods = ["BBVI"]
        self.default_method = "BBVI"
        self.multivariate_model = False

        # Format the data
        self.is_pandas = True # This is compulsory for this model type
        self.data_original = data.copy()
        self.formula = formula
        self.y, self.X = dmatrices(formula, data)
        self.y_name = self.y.design_info.describe()
        self.X_names = self.X.design_info.describe().split(" + ")
        self.y = self.y.astype(np.float) 
        self.X = self.X.astype(np.float) 
        self.z_no = self.X.shape[1]
        self.data_name = self.y_name
        self.y = np.array([self.y]).ravel()
        self.data = self.y.copy()
        self.X = np.array([self.X])[0]
        self.index = data.index
        self.data_length = self.data.shape[0]
        self.X = self.X[self.ar:, :]
        self.X = np.concatenate([self._ar_matrix().T, self.X], axis=1).T

        self._create_latent_variables()

        self.family = family
        
        self.model_name2, self.link, self.scale, self.shape, self.skewness, self.mean_transform, self.cythonized = self.family.setup()
        
        self.model_name = self.model_name2 + " NNARX(" + str(self.ar) + ")"

        # Build any remaining latent variables that are specific to the family chosen
        for no, i in enumerate(self.family.build_latent_variables()):
            self.latent_variables.add_z(i[0], i[1], i[2])
            self.latent_variables.z_list[-1].start = i[3]

        self.z_no = len(self.latent_variables.z_list)
        self.family_z_no = len(self.family.build_latent_variables())

        # Initialize with random weights
        for var_no in range(len(self.latent_variables.z_list)-self.family_z_no):
            self.latent_variables.z_list[var_no].start = np.random.normal()

        if isinstance(self.family, fam.Normal):
            self.neg_loglik = self.normal_neg_loglik
        else:
            self.neg_loglik = self.general_neg_loglik

    def _ar_matrix(self):
        """ Creates Autoregressive matrix

        Returns
        ----------
        X : np.ndarray
            Autoregressive Matrix

        """

        if self.ar != 0:
            X = self.data[(self.max_lag-1):-1]
            for i in range(1, self.ar):
                X = np.vstack((X, self.data[(self.max_lag-i-1):-i-1]))
            return X
        else:
            return np.zeros(self.data_length-self.max_lag)
            
    def _create_latent_variables(self):
        """ Creates the model's latent variables

        Returns
        ----------
        None (changes model attributes)
        """

        # Input layer
        for unit in range(self.units):
            self.latent_variables.add_z('Constant | Layer ' + str(1) + ' | Unit ' + str(unit+1), fam.Cauchy(0,1,transform=None), fam.Normal(0, 3))

            for ar_term in range(self.ar):
                self.latent_variables.add_z('AR' + str(ar_term+1) + ' | Layer ' + str(1) + ' | Unit ' + str(unit+1), fam.Cauchy(0,1,transform=None), fam.Normal(0, 3))

            for z in range(len(self.X_names)):
                self.latent_variables.add_z('Weight ' + self.X_names[z], fam.Cauchy(0, 1, transform=None), fam.Normal(0, 3))

        # Hidden layers
        for layer in range(1, self.layers):
            for unit in range(self.units):
                for weight in range(self.units):
                    self.latent_variables.add_z('Weight ' + str(weight+1) + ' | Layer ' + str(layer+1) + ' | Unit ' + str(unit+1), fam.Cauchy(0,1,transform=None), fam.Normal(0, 3))

        # Output layer
        for weight in range(self.units):
            self.latent_variables.add_z('Output Weight ' + str(weight+1), fam.Cauchy(0,1,transform=None), fam.Normal(0, 3))

    def _get_scale_and_shape(self,parm):
        """ Obtains appropriate model scale and shape latent variables

        Parameters
        ----------
        parm : np.array
            Transformed latent variable vector

        Returns
        ----------
        None (changes model attributes)
        """

        if self.scale is True:
            if self.shape is True:
                model_shape = parm[-1]  
                model_scale = parm[-2]
            else:
                model_shape = 0
                model_scale = parm[-1]
        else:
            model_scale = 0
            model_shape = 0 

        if self.skewness is True:
            model_skewness = parm[-3]
        else:
            model_skewness = 0

        return model_scale, model_shape, model_skewness


    def _get_scale_and_shape_sim(self, transformed_lvs):
        """ Obtains model scale, shape, skewness latent variables for
        a 2d array of simulations.

        Parameters
        ----------
        transformed_lvs : np.array
            Transformed latent variable vector (2d - with draws of each variable)

        Returns
        ----------
        - Tuple of np.arrays (each being scale, shape and skewness draws)
        """

        if self.scale is True:
            if self.shape is True:
                model_shape = self.latent_variables.z_list[-1].prior.transform(transformed_lvs[-1, :]) 
                model_scale = self.latent_variables.z_list[-2].prior.transform(transformed_lvs[-2, :])
            else:
                model_shape = np.zeros(transformed_lvs.shape[1])
                model_scale = self.latent_variables.z_list[-1].prior.transform(transformed_lvs[-1, :])
        else:
            model_scale = np.zeros(transformed_lvs.shape[1])
            model_shape = np.zeros(transformed_lvs.shape[1])

        if self.skewness is True:
            model_skewness = self.latent_variables.z_list[-3].prior.transform(transformed_lvs[-3, :])
        else:
            model_skewness = np.zeros(transformed_lvs.shape[1])

        return model_scale, model_shape, model_skewness

    def _model(self, beta):
        """ Creates the structure of the model (model matrices etc)

        Parameters
        ----------
        beta : np.ndarray
            Contains untransformed starting values for the latent variables

        Returns
        ----------
        mu : np.ndarray
            Contains the predicted values (location) for the time series

        Y : np.ndarray
            Contains the length-adjusted time series (accounting for lags)
        """     


        Y = np.array(self.data[self.max_lag:])

        # Transform latent variables
        z = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])

        return neural_network_tanh(Y, self.X, z, self.units, self.layers, self.ar+len(self.X_names)), Y

    def _mb_model(self, beta, mini_batch):
        """ Creates the structure of the model (model matrices etc) for mini batch model

        Parameters
        ----------
        beta : np.ndarray
            Contains untransformed starting values for the latent variables

        mini_batch : int
            Mini batch size for the data sampling

        Returns
        ----------
        mu : np.ndarray
            Contains the predicted values (location) for the time series

        Y : np.ndarray
            Contains the length-adjusted time series (accounting for lags)
        """     

        Y = np.array(self.data[self.max_lag:])

        sample = np.random.choice(len(Y), mini_batch, replace=False)

        Y = Y[sample]
        X = self.X[:, sample]

        # Transform latent variables
        z = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])

        return neural_network_tanh_mb(Y, X, z, self.units, self.layers, self.ar+len(self.X_names)), Y

    def predict_new(self, X, z):

        first_layer_output = np.zeros(self.units)
        
        for unit in range(self.units):
            first_layer_output[unit] = self.activation(np.matmul(np.transpose(X), z[unit*(self.ar+len(self.X_names)+1):((unit+1)*(self.ar+len(self.X_names)+1))]))

        params_used = ((self.units)*(self.ar+len(self.X_names)+1))

        # Hidden layers
        hidden_layer_output = np.zeros((self.units, self.layers-1))
        for layer in range(1, self.layers):
            for unit in range(self.units):
                if layer == 1:
                    hidden_layer_output[unit,layer-1] = self.activation(np.matmul(first_layer_output,
                        z[params_used+unit*(self.units)+((layer-1)*self.units**2):((params_used+(unit+1)*self.units)+((layer-1)*self.units**2))]))
                else:
                    hidden_layer_output[unit,layer-1] = self.activation(np.matmul(hidden_layer_output[:,layer-1],
                        z[params_used+unit*(self.units)+((layer-1)*self.units**2):((params_used+(unit+1)*self.units)+((layer-1)*self.units**2))]))

        params_used = params_used + (self.layers-1)*self.units**2

        # Output layer
        if self.layers == 1:
            mu = np.matmul(first_layer_output, z[params_used:params_used+self.units])
        else:
            mu = np.matmul(hidden_layer_output[:,-1], z[params_used:params_used+self.units])

        return mu

    def _mean_prediction(self, mu, Y, h, t_z):
        """ Creates a h-step ahead mean prediction

        Parameters
        ----------
        mu : np.ndarray
            The past predicted values

        Y : np.ndarray
            The past data

        h : int
            How many steps ahead for the prediction

        t_z : np.ndarray
            A vector of (transformed) latent variables

        Returns
        ----------
        h-length vector of mean predictions
        """     

        # Create arrays to iteratre over
        Y_exp = Y.copy()

        # Loop over h time periods          
        for t in range(0,h):

            if self.ar != 0:
                Y_exp_normalized = (Y_exp[-self.ar:][::-1] - self._norm_mean) / self._norm_std
                new_value = self.predict_new(np.append(1.0, Y_exp_normalized), self.latent_variables.get_z_values())

            else:  
                new_value = self.predict_new(np.array([1.0]), self.latent_variables.get_z_values())

            Y_exp = np.append(Y_exp, [self.link(new_value)])

        return Y_exp

    def _sim_prediction(self, mu, Y, h, t_z, simulations):
        """ Simulates a h-step ahead mean prediction

        Parameters
        ----------
        mu : np.ndarray
            The past predicted values

        Y : np.ndarray
            The past data

        h : int
            How many steps ahead for the prediction

        t_z : np.ndarray
            A vector of (transformed) latent variables

        simulations : int
            How many simulations to perform

        Returns
        ----------
        Matrix of simulations
        """     

        model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_z)
        sim_vector = np.zeros([simulations,h])

        for n in range(0, simulations):
            # Create arrays to iteratre over        
            Y_exp = Y.copy()

            # Loop over h time periods          
            for t in range(0,h):

                if self.ar != 0:
                    Y_exp_normalized = (Y_exp[-self.ar:][::-1] - self._norm_mean) / self._norm_std
                    new_value = self.predict_new(np.append(1.0, Y_exp_normalized), self.latent_variables.get_z_values())

                else:
                    new_value = self.predict_new(np.array([1.0]), self.latent_variables.get_z_values())

                new_value += np.random.randn(1)*t_z[-1]

                if self.model_name2 == "Exponential":
                    rnd_value = self.family.draw_variable(1.0/self.link(new_value), model_scale, model_shape, model_skewness, 1)[0]
                else:
                    rnd_value = self.family.draw_variable(self.link(new_value), model_scale, model_shape, model_skewness, 1)[0]

                Y_exp = np.append(Y_exp, [rnd_value])

            sim_vector[n] = Y_exp[-h:]

        return np.transpose(sim_vector)

    def _summarize_simulations(self, mean_values, sim_vector, date_index, h, past_values):
        """ Produces simulation forecasted values and prediction intervals

        Parameters
        ----------
        mean_values : np.ndarray
            Mean predictions for h-step ahead forecasts

        sim_vector : np.ndarray
            N simulated predictions for h-step ahead forecasts

        date_index : pd.DateIndex or np.ndarray
            Date index for the simulations

        h : int
            How many steps ahead are forecast

        past_values : int
            How many past observations to include in the forecast plot

        intervals : boolean
            Would you like to show prediction intervals for the forecast?
        """         

        error_bars = []
        for pre in range(5,100,5):
            error_bars.append(np.insert([np.percentile(i,pre) for i in sim_vector] - mean_values[-h:],0,0))
        forecasted_values = mean_values[-h-1:]
        plot_values = mean_values[-h-past_values:]
        plot_index = date_index[-h-past_values:]
        return error_bars, forecasted_values, plot_values, plot_index
        
    def general_neg_loglik(self, beta):
        """ Calculates the negative log-likelihood of the model

        Parameters
        ----------
        beta : np.ndarray
            Contains untransformed starting values for latent variables

        Returns
        ----------
        The negative logliklihood of the model
        """     

        mu, Y = self._model(beta)
        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        #TODO: Replace above with transformation that only acts on scale, shape, skewness in future (speed-up)
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)
        return self.family.neg_loglikelihood(Y, self.link(mu), model_scale, model_shape, model_skewness)

    def mb_neg_loglik(self, beta, mini_batch):
        """ Calculates the negative log-likelihood of the model for a minibatch

        Parameters
        ----------
        beta : np.ndarray
            Contains untransformed starting values for latent variables

        mini_batch : int
            Size of each mini batch of data

        Returns
        ----------
        The negative logliklihood of the model
        """     

        mu, Y = self._mb_model(beta, mini_batch)
        parm = np.array([self.latent_variables.z_list[k].prior.transform(beta[k]) for k in range(beta.shape[0])])
        #TODO: Replace above with transformation that only acts on scale, shape, skewness in future (speed-up)
        model_scale, model_shape, model_skewness = self._get_scale_and_shape(parm)
        return self.family.neg_loglikelihood(Y, self.link(mu), model_scale, model_shape, model_skewness)

    def normal_neg_loglik(self, beta):
        """ Creates the negative log-likelihood of the model

        Parameters
        ----------
        beta : np.ndarray
            Contains untransformed starting values for latent variables

        Returns
        ----------
        The negative logliklihood of the model
        """     

        mu, Y = self._model(beta)
        return -np.sum(ss.norm.logpdf(Y, loc=mu, scale=self.latent_variables.z_list[-1].prior.transform(beta[-1])))

    def plot_fit(self, **kwargs):
        """ 
        Plots the fit of the model against the data
        """

        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))
        plt.figure(figsize=figsize)
        date_index = self.index[self.ar:self.data.shape[0]]
        mu, Y = self._model(self.latent_variables.get_z_values())
        plt.plot(date_index,Y,label='Data')
        plt.plot(date_index,mu,label='Filter',c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def plot_predict(self, h=5, past_values=20, intervals=True, **kwargs):
        """ Plots forecasts with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : boolean
            Would you like to show prediction intervals for the forecast?

        Returns
        ----------
        - Plot of the forecast
        """     

        figsize = kwargs.get('figsize',(10,7))

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:

            import matplotlib.pyplot as plt
            import seaborn as sns

            # Retrieve data, dates and (transformed) latent variables
            mu, Y = self._model(self.latent_variables.get_z_values())         
            date_index = self.shift_dates(h)
            t_z = self.transform_z()

            # Get mean prediction and simulations (for errors)
            mean_values = self._mean_prediction(mu, Y, h, t_z)
            if intervals is True:
                sim_values = self._sim_prediction(mu, Y, h, t_z, 15000)
            else:
                sim_values = self._sim_prediction(mu, Y, h, t_z, 2)
            error_bars, forecasted_values, plot_values, plot_index = self._summarize_simulations(mean_values, sim_values, date_index, h, past_values)

            plt.figure(figsize=figsize)
            if intervals is True:
                alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                for count, pre in enumerate(error_bars):
                    plt.fill_between(date_index[-h-1:], forecasted_values-pre, forecasted_values+pre,alpha=alpha[count])            
            plt.plot(plot_index,plot_values)
            plt.title("Forecast for " + self.data_name)
            plt.xlabel("Time")
            plt.ylabel(self.data_name)
            plt.show()

    def predict_is(self, h=5, fit_once=True, fit_method='MLE', intervals=False, **kwargs):
        """ Makes dynamic out-of-sample predictions with the estimated model on in-sample data

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        fit_once : boolean
            (default: True) Fits only once before the in-sample prediction; if False, fits after every new datapoint

        fit_method : string
            Which method to fit the model with

        intervals: boolean
            Whether to return prediction intervals

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     
        predictions = []

        for t in range(0,h):
            x = NNAR(ar=self.ar, units=self.units, 
                layers=self.layers, data=self.data_original[:-h+t], family=self.family)
            if fit_once is False:
                x.fit(method=fit_method, printer=False)
            if t == 0:
                if fit_once is True:
                    x.fit(method=fit_method, printer=False)
                    saved_lvs = x.latent_variables
                predictions = x.predict(1, intervals=intervals)
            else:
                if fit_once is True:
                    x.latent_variables = saved_lvs
                predictions = pd.concat([predictions,x.predict(1, intervals=intervals)])
        
        if intervals is True:
            predictions.rename(columns={0:self.data_name, 1: "1% Prediction Interval", 
                2: "5% Prediction Interval", 3: "95% Prediction Interval", 4: "99% Prediction Interval"}, inplace=True)
        else:
            predictions.rename(columns={0:self.data_name}, inplace=True)

        predictions.index = self.index[-h:]

        return predictions

    def plot_predict_is(self, h=5, fit_once=True, fit_method='MLE', **kwargs):
        """ Plots forecasts with the estimated model against data (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        fit_once : boolean
            (default: True) Fits only once before the in-sample prediction; if False, fits after every new datapoint

        fit_method : string
            Which method to fit the model with

        Returns
        ----------
        - Plot of the forecast against data 
        """     

        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))
        plt.figure(figsize=figsize)
        predictions = self.predict_is(h, fit_method=fit_method, fit_once=fit_once)
        data = self.data[-h:]
        plt.plot(predictions.index, data, label='Data')
        plt.plot(predictions.index, predictions, label='Predictions', c='black')
        plt.title(self.data_name)
        plt.legend(loc=2)   
        plt.show()          

    def predict(self, h=5, intervals=False):
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        intervals : boolean (default: False)
            Whether to return prediction intervals

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:

            mu, Y = self._model(self.latent_variables.get_z_values())   
            date_index = self.shift_dates(h)

            if self.latent_variables.estimation_method in ['M-H']:
                sim_vector = self._sim_prediction_bayes(h, 15000)

                forecasted_values = np.array([np.mean(i) for i in sim_vector])
                prediction_01 = np.array([np.percentile(i, 1) for i in sim_vector])
                prediction_05 = np.array([np.percentile(i, 5) for i in sim_vector])
                prediction_95 = np.array([np.percentile(i, 95) for i in sim_vector])
                prediction_99 = np.array([np.percentile(i, 99) for i in sim_vector])

            else:
                t_z = self.transform_z()
                mean_values = self._mean_prediction(mu, Y, h, t_z)
                if intervals is True:
                    sim_values = self._sim_prediction(mu, Y, h, t_z, 15000)
                else:
                    sim_values = self._sim_prediction(mu, Y, h, t_z, 2)

                if self.model_name2 == "Skewt":
                    model_scale, model_shape, model_skewness = self._get_scale_and_shape(t_z)
                    m1 = (np.sqrt(model_shape)*sp.gamma((model_shape-1.0)/2.0))/(np.sqrt(np.pi)*sp.gamma(model_shape/2.0))
                    forecasted_values = mean_values[-h:] + (model_skewness - (1.0/model_skewness))*model_scale*m1 
                else:
                    forecasted_values = mean_values[-h:] 

            if intervals is False:
                result = pd.DataFrame(forecasted_values)
                result.rename(columns={0:self.data_name}, inplace=True)
            else:
                # Get mean prediction and simulations (for errors)
                if self.latent_variables.estimation_method not in ['M-H']:
                    sim_values = self._sim_prediction(mu, Y, h, t_z, 15000)
                    prediction_01 = np.array([np.percentile(i, 1) for i in sim_values])
                    prediction_05 = np.array([np.percentile(i, 5) for i in sim_values])
                    prediction_95 = np.array([np.percentile(i, 95) for i in sim_values])
                    prediction_99 = np.array([np.percentile(i, 99) for i in sim_values])

                result = pd.DataFrame([forecasted_values, prediction_01, prediction_05, 
                    prediction_95, prediction_99]).T
                result.rename(columns={0:self.data_name, 1: "1% Prediction Interval", 
                    2: "5% Prediction Interval", 3: "95% Prediction Interval", 4: "99% Prediction Interval"}, 
                    inplace=True)
 
            result.index = date_index[-h:]

            return result

    def sample(self, nsims=1000):
        """ Samples from the posterior predictive distribution

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws from the posterior predictive distribution

        Returns
        ----------
        - np.ndarray of draws from the data
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            lv_draws = self.draw_latent_variables(nsims=nsims)
            mus = [self._model(lv_draws[:,i])[0] for i in range(nsims)]
            model_scale, model_shape, model_skewness = self._get_scale_and_shape_sim(lv_draws)
            data_draws = np.array([self.family.draw_variable(self.link(mus[i]), 
                np.repeat(model_scale[i], mus[i].shape[0]), np.repeat(model_shape[i], mus[i].shape[0]), 
                np.repeat(model_skewness[i], mus[i].shape[0]), mus[i].shape[0]) for i in range(nsims)])
            return data_draws

    def plot_sample(self, nsims=10, plot_data=True, **kwargs):
        """
        Plots draws from the posterior predictive density against the data

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws from the posterior predictive distribution

        plot_data boolean
            Whether to plot the data or not
        """

        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            import matplotlib.pyplot as plt
            import seaborn as sns
            
            figsize = kwargs.get('figsize',(10,7))
            plt.figure(figsize=figsize)
            date_index = self.index[self.ar:self.data_length]
            mu, Y = self._model(self.latent_variables.get_z_values())
            draws = self.sample(nsims).T
            plt.plot(date_index, draws, label='Posterior Draws', alpha=1.0)
            if plot_data is True:
                plt.plot(date_index, Y, label='Data', c='black', alpha=0.5, linestyle='', marker='s')
            plt.title(self.data_name)
            plt.show()    

    def ppc(self, nsims=1000, T=np.mean):
        """ Computes posterior predictive p-value

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws for the PPC

        T : function
            A discrepancy measure - e.g. np.mean, np.std, np.max

        Returns
        ----------
        - float (posterior predictive p-value)
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            lv_draws = self.draw_latent_variables(nsims=nsims)
            mus = [self._model(lv_draws[:,i])[0] for i in range(nsims)]
            model_scale, model_shape, model_skewness = self._get_scale_and_shape_sim(lv_draws)
            data_draws = np.array([self.family.draw_variable(self.link(mus[i]), 
                np.repeat(model_scale[i], mus[i].shape[0]), np.repeat(model_shape[i], mus[i].shape[0]), 
                np.repeat(model_skewness[i], mus[i].shape[0]), mus[i].shape[0]) for i in range(nsims)])
            T_sims = T(self.sample(nsims=nsims), axis=1)
            T_actual = T(self.data)
            return len(T_sims[T_sims>T_actual])/nsims

    def plot_ppc(self, nsims=1000, T=np.mean, **kwargs):
        """ Plots histogram of the discrepancy from draws of the posterior

        Parameters
        ----------
        nsims : int (default : 1000)
            How many draws for the PPC

        T : function
            A discrepancy measure - e.g. np.mean, np.std, np.max
        """     
        if self.latent_variables.estimation_method not in ['BBVI', 'M-H']:
            raise Exception("No latent variables estimated!")
        else:
            import matplotlib.pyplot as plt
            import seaborn as sns
            
            figsize = kwargs.get('figsize',(10,7))

            lv_draws = self.draw_latent_variables(nsims=nsims)
            mus = [self._model(lv_draws[:,i])[0] for i in range(nsims)]
            model_scale, model_shape, model_skewness = self._get_scale_and_shape_sim(lv_draws)
            data_draws = np.array([self.family.draw_variable(self.link(mus[i]), 
                np.repeat(model_scale[i], mus[i].shape[0]), np.repeat(model_shape[i], mus[i].shape[0]), 
                np.repeat(model_skewness[i], mus[i].shape[0]), mus[i].shape[0]) for i in range(nsims)])
            T_sim = T(self.sample(nsims=nsims), axis=1)
            T_actual = T(self.data)

            if T == np.mean:
                description = " of the mean"
            elif T == np.max:
                description = " of the maximum"
            elif T == np.min:
                description = " of the minimum"
            elif T == np.median:
                description = " of the median"
            else:
                description = ""

            plt.figure(figsize=figsize)
            ax = plt.subplot()
            ax.axvline(T_actual)
            sns.distplot(T_sim, kde=False, ax=ax)
            ax.set(title='Posterior predictive' + description, xlabel='T(x)', ylabel='Frequency');
            plt.show()


import numpy as np
import scipy.stats as ss
import scipy.special as sp

from .family import Family
from .flat import Flat

from .gas_recursions import gas_recursion_normal_orderone, gas_recursion_normal_ordertwo
from .gas_recursions import gasx_recursion_normal_orderone, gasx_recursion_normal_ordertwo
from .gas_recursions import gas_llev_recursion_normal_orderone, gas_llev_recursion_normal_ordertwo
from .gas_recursions import gas_llt_recursion_normal_orderone, gas_llt_recursion_normal_ordertwo
from .gas_recursions import gas_reg_recursion_normal_orderone, gas_reg_recursion_normal_ordertwo


class Normal(Family):
    """ 
    Normal Distribution
    ----
    This class contains methods relating to the normal distribution for time series.
    """

    def __init__(self, mu=0.0, sigma=1.0, transform=None, **kwargs):
        """
        Parameters
        ----------
        mu : float
            Mean parameter for the Normal distribution

        sigma : float
            Standard deviation for the Normal distribution

        transform : str
            Whether to apply a transformation for the location latent variable - e.g. 'exp' or 'logit'
        """
        super(Normal, self).__init__(transform)
        self.mu0 = mu
        self.sigma0 = sigma
        self.param_no = 2
        self.covariance_prior = False

        self.gradient_only = kwargs.get('gradient_only', False) # used for GAS Normal models
        if self.gradient_only is True:
            self.score_function = self.first_order_score
        else:
            self.score_function = self.second_order_score

    def approximating_model(self, beta, T, Z, R, Q, h_approx, data):
        """ Creates approximating Gaussian state space model for the Normal measurement density
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm
        
        h_approx : float
            The variance of the measurement density
        
        data: np.array
            The univariate time series data

        Returns
        ----------
        H : np.array
            Approximating measurement variance matrix
        
        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(data.shape[0])*h_approx
        mu = np.zeros(data.shape[0])

        return H, mu

    def approximating_model_reg(self, beta, T, Z, R, Q, h_approx, data, X, state_no):
        """ Creates approximating Gaussian state space model for the Normal measurement density
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm
        
        h_approx : float
            The variance of the measurement density
        
        data: np.array
            The univariate time series data

        X: np.array
            The regressors

        state_no : int
            Number of states
    
        Returns
        ----------
        H : np.array
            Approximating measurement variance matrix
        
        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(data.shape[0])*h_approx
        mu = np.zeros(data.shape[0])

        return H, mu

    @staticmethod
    def build_latent_variables():
        """ Builds additional latent variables for this family in a probabilistic model

        Returns
        ----------
        - A list of lists (each sub-list contains latent variable information)
        """
        lvs_to_build = []
        lvs_to_build.append(['Normal Scale', Flat(transform='exp'), Normal(0, 3), 0.0])
        return lvs_to_build

    @staticmethod
    def draw_variable(loc, scale, shape, skewness, nsims):
        """ Draws random variables from this distribution with new latent variables

        Parameters
        ----------
        loc : float
            location parameter for the distribution

        scale : float
            scale parameter for the distribution

        shape : float
            tail thickness parameter for the distribution

        skewness : float
            skewness parameter for the distribution

        nsims : int or list
            number of draws to take from the distribution

        Returns
        ----------
        - Random draws from the distribution
        """
        return np.random.normal(loc, scale, nsims)

    def draw_variable_local(self, size): 
        """ Simulate from the Normal distribution using instance values

        Parameters
        ----------
        size : int
            How many simulations to perform

        Returns
        ----------
        np.ndarray of Normal random variable
        """
        return ss.norm.rvs(loc=self.mu0, scale=self.sigma0, size=size)

    @staticmethod
    def first_order_score(y, mean, scale, shape, skewness):
        """ GAS Normal Update term using gradient only - native Python function

        Parameters
        ----------
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Normal distribution

        scale : float
            scale parameter for the Normal distribution

        shape : float
            tail thickness parameter for the Normal distribution

        skewness : float
            skewness parameter for the Normal distribution

        Returns
        ----------
        - Score of the Normal family
        """
        return (y-mean)/np.power(scale,2)

    def logpdf(self, mu):
        """
        Log PDF for Normal prior

        Parameters
        ----------
        mu : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - log(p(mu))
        """
        if self.transform is not None:
            mu = self.transform(mu)     
        return -np.log(float(self.sigma0)) - (0.5*(mu-self.mu0)**2)/float(self.sigma0**2)

    @staticmethod
    def markov_blanket(y, mean, scale, shape, skewness):
        """ Markov blanket for each likelihood term - used for state space models

        Parameters
        ----------
        y : np.ndarray
            univariate time series

        mean : np.ndarray
            array of location parameters for the Normal distribution

        scale : float
            scale parameter for the Normal distribution

        shape : float
            tail thickness parameter for the Normal distribution

        skewness : float
            skewness parameter for the Normal distribution

        Returns
        ----------
        - Markov blanket of the Normal family
        """
        return ss.norm.logpdf(y, loc=mean, scale=scale)

    @staticmethod
    def setup():
        """ Returns the attributes of this family if using in a probabilistic model

        Notes
        ----------
        - scale notes whether family has a variance parameter (sigma)
        - shape notes whether family has a tail thickness parameter (nu)
        - skewness notes whether family has a skewness parameter (gamma)
        - mean_transform is a function which transforms the location parameter
        - cythonized notes whether the family has cythonized routines
        
        Returns
        ----------
        - model name, link function, scale, shape, skewness, mean_transform, cythonized
        """
        name = "Normal"
        link = np.array
        scale = True
        shape = False
        skewness = False
        mean_transform = np.array
        cythonized = True
        return name, link, scale, shape, skewness, mean_transform, cythonized

    @staticmethod
    def neg_loglikelihood(y, mean, scale, shape, skewness):
        """ Negative loglikelihood function for this distribution

        Parameters
        ----------
        y : np.ndarray
            univariate time series

        mean : np.ndarray
            array of location parameters for the Normal distribution

        scale : float
            scale parameter for the Normal distribution

        shape : float
            tail thickness parameter for the Normal distribution

        skewness : float
            skewness parameter for the Normal distribution

        Returns
        ----------
        - Negative loglikelihood of the Normal family
        """
        return -np.sum(ss.norm.logpdf(y, loc=mean, scale=scale))

    def pdf(self, mu):
        """
        PDF for Normal prior

        Parameters
        ----------
        mu : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - p(mu)
        """
        if self.transform is not None:
            mu = self.transform(mu)             
        return (1.0/float(self.sigma0))*np.exp(-(0.5*(mu-self.mu0)**2)/float(self.sigma0**2))

    @staticmethod
    def reg_score_function(X, y, mean, scale, shape, skewness):
        """ GAS Normal Regression Update term using gradient only - native Python function

        Parameters
        ----------
        X : float
            datapoint for the right hand side variable
    
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Normal distribution

        scale : float
            scale parameter for the Normal distribution

        shape : float
            tail thickness parameter for the Normal distribution

        skewness : float
            skewness parameter for the Normal distribution

        Returns
        ----------
        - Score of the Normal family
        """
        return X*(y-mean)

    @staticmethod
    def second_order_score(y, mean, scale, shape, skewness):
        """ GAS Normal Update term potentially using second-order information - native Python function

        Parameters
        ----------
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Normal distribution

        scale : float
            scale parameter for the Normal distribution

        shape : float
            tail thickness parameter for the Normal distribution

        skewness : float
            skewness parameter for the Normal distribution

        Returns
        ----------
        - Adjusted score of the Normal family
        """
        return y-mean

    def vi_change_param(self, index, value):
        """ Wrapper function for changing latent variables - variational inference

        Parameters
        ----------
        index : int
            0 or 1 depending on which latent variable

        value : float
            What to change the latent variable to
        """
        if index == 0:
            self.mu0 = value
        elif index == 1:
            self.sigma0 = np.exp(value)

    def vi_return_param(self, index):
        """ Wrapper function for selecting appropriate latent variable for variational inference

        Parameters
        ----------
        index : int
            0 or 1 depending on which latent variable

        Returns
        ----------
        The appropriate indexed parameter
        """
        if index == 0:
            return self.mu0
        elif index == 1:
            return np.log(self.sigma0)

    def vi_loc_score(self,x):
        """ The gradient of the location latent variable mu - used for variational inference

        Parameters
        ----------
        x : float
            A random variable

        Returns
        ----------
        The gradient of the location latent variable mu at x
        """
        return (x-self.mu0)/(self.sigma0**2)

    def vi_scale_score(self,x):
        """ The score of the scale, where scale = exp(x) - used for variational inference

        Parameters
        ----------
        x : float
            A random variable

        Returns
        ----------
        The gradient of the scale latent variable at x
        """
        return np.exp(-2.0*np.log(self.sigma0))*(x-self.mu0)**2 - 1.0

    def vi_score(self, x, index):
        """ Wrapper function for selecting appropriate score

        Parameters
        ----------
        x : float
            A random variable

        index : int
            0 or 1 depending on which latent variable

        Returns
        ----------
        The gradient of the scale latent variable at x
        """
        if index == 0:
            return self.vi_loc_score(x)
        elif index == 1:
            return self.vi_scale_score(x)


    # Optional Cythonized recursions below for GAS Normal models

    @staticmethod
    def gradient_recursion():
        """ GAS Normal Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Normal model - gradient only
        """
        return gas_recursion_normal_orderone

    @staticmethod
    def newton_recursion():
        """ GAS Normal Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Normal model - adjusted score
        """
        return gas_recursion_normal_ordertwo

    @staticmethod
    def gradientx_recursion():
        """ GASX Normal Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GASX Normal model - gradient only
        """
        return gasx_recursion_normal_orderone

    @staticmethod
    def newtonx_recursion():
        """ GASX Normal Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GASX Normal model - adjusted score
        """
        return gasx_recursion_normal_ordertwo

    @staticmethod
    def gradientllev_recursion():
        """ GAS Local Level Normal Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Local Level Normal model - gradient only
        """
        return gas_llev_recursion_normal_orderone

    @staticmethod
    def newtonllev_recursion():
        """ GAS Local Level Normal Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Local Level Normal model - adjusted score
        """
        return gas_llev_recursion_normal_ordertwo

    @staticmethod
    def gradientllt_recursion():
        """ GAS Local Linear Trend Normal Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Local Linear Trend Normal model - gradient only
        """
        return gas_llt_recursion_normal_orderone

    @staticmethod
    def newtonllt_recursion():
        """ GAS Local Linear Trend Normal Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Local Linear Trend Normal model - adjusted score
        """
        return gas_llt_recursion_normal_ordertwo

    @staticmethod
    def gradientreg_recursion():
        """ GAS Dynamic Regression Normal Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Dynamic Regression Normal model - gradient only
        """
        return gas_reg_recursion_normal_orderone

    @staticmethod
    def newtonreg_recursion():
        """ GAS Dynamic Regression Normal Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Dynamic Regression Normal model - adjusted score
        """
        return gas_reg_recursion_normal_ordertwo


import numpy as np
import scipy.stats as ss
import scipy.special as sp

from .poisson_kalman_recursions import nl_univariate_KFS, nld_univariate_KFS
from .family import Family
from .flat import Flat
from .normal import Normal

from .gas_recursions import gas_recursion_poisson_orderone, gas_recursion_poisson_ordertwo
from .gas_recursions import gasx_recursion_poisson_orderone, gasx_recursion_poisson_ordertwo
from .gas_recursions import gas_llev_recursion_poisson_orderone, gas_llev_recursion_poisson_ordertwo
from .gas_recursions import gas_llt_recursion_poisson_orderone, gas_llt_recursion_poisson_ordertwo
from .gas_recursions import gas_reg_recursion_poisson_orderone, gas_reg_recursion_poisson_ordertwo


class Poisson(Family):
    """ 
    Poisson Distribution
    ----
    This class contains methods relating to the Poisson distribution for time series.
    """

    def __init__(self, lmd=1.0, transform=None, **kwargs):
        """
        Parameters
        ----------
        lambda : float
            Rate parameter for the Poisson distribution

        transform : str
            Whether to apply a transformation to the Poisson latent variable - e.g. 'exp' or 'logit'
        """
        super(Poisson, self).__init__(transform)
        self.lmd0 = lmd
        self.covariance_prior = False

        self.gradient_only = kwargs.get('gradient_only', False) # used for GAS Poisson models
        if self.gradient_only is True:
            self.score_function = self.first_order_score
        else:
            self.score_function = self.second_order_score

    def approximating_model(self, beta, T, Z, R, Q, h_approx, data):
        """ Creates approximating Gaussian state space model for Poisson measurement density
        
        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables
        
        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm
        
        h_approx : float
            The variance of the measurement density
        
        data: np.array
            The univariate time series data

        Returns
        ----------
        H : np.array
            Approximating measurement variance matrix
        
        mu : np.array
            Approximating measurement constants
        """     

        H = np.ones(data.shape[0])
        mu = np.zeros(data.shape[0])

        alpha = np.array([np.zeros(data.shape[0])])
        tol = 100.0
        it = 0
        while tol > 10**-7 and it < 5:
            old_alpha = alpha[0]
            alpha, V = nl_univariate_KFS(data,Z,H,T,Q,R,mu)
            H = np.exp(-alpha[0])
            mu = data - alpha[0] - np.exp(-alpha[0])*(data - np.exp(alpha[0]))
            tol = np.mean(np.abs(alpha[0]-old_alpha))
            it += 1

        return H, mu

    def approximating_model_reg(self, beta, T, Z, R, Q, h_approx, data, X, state_no):
        """ Creates approximating Gaussian model for Poisson measurement density
        - dynamic regression model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        T, Z, R, Q : np.array
            State space matrices used in KFS algorithm

        data: np.array
            The univariate time series data

        X: np.array
            The regressors

        state_no : int
            Number of states

        Returns
        ----------

        H : np.array
            Approximating measurement variance matrix

        mu : np.array
            Approximating measurement constants
        """     
        
        H = np.ones(data.shape[0])
        mu = np.zeros(data.shape[0])

        alpha = np.zeros([state_no, data.shape[0]])
        tol = 100.0
        it = 0
        while tol > 10**-7 and it < 5:
            old_alpha = np.sum(X*alpha.T,axis=1)
            alpha, V = nld_univariate_KFS(data,Z,H,T,Q,R,mu)
            H = np.exp(-np.sum(X*alpha.T,axis=1))
            mu = data - np.sum(X*alpha.T,axis=1) - np.exp(-np.sum(X*alpha.T,axis=1))*(data - np.exp(np.sum(X*alpha.T,axis=1)))
            tol = np.mean(np.abs(np.sum(X*alpha.T,axis=1)-old_alpha))
            it += 1

        return H, mu

    @staticmethod
    def build_latent_variables():
        """ Builds additional latent variables for this family in a probabilistic model

        Returns
        ----------
        - A list of lists (each sub-list contains latent variable information)
        """
        lvs_to_build = []
        return lvs_to_build

    @staticmethod
    def draw_variable(loc, scale, shape, skewness, nsims):
        """ Draws random variables from Poisson distribution

        Parameters
        ----------
        loc : float
            location parameter for the distribution

        scale : float
            scale parameter for the distribution

        shape : float
            tail thickness parameter for the distribution

        skewness : float
            skewness parameter for the distribution

        nsims : int or list
            number of draws to take from the distribution

        Returns
        ----------
        - Random draws from the distribution
        """
        return np.random.poisson(loc, nsims)

    @staticmethod
    def first_order_score(y, mean, scale, shape, skewness):
        """ GAS Poisson Update term using gradient only - native Python function

        Parameters
        ----------
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Poisson distribution

        scale : float
            scale parameter for the Poisson distribution

        shape : float
            tail thickness parameter for the Poisson distribution

        skewness : float
            skewness parameter for the Poisson distribution

        Returns
        ----------
        - Score of the Poisson family
        """
        return y-mean

    def logpdf(self, mu):
        """
        Log PDF for Poisson prior

        Parameters
        ----------
        mu : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - log(p(mu))
        """
        if self.transform is not None:
            mu = self.transform(mu)    
        return ss.poisson.logpmf(mu, self.lmd0) 

    @staticmethod
    def markov_blanket(y, mean, scale, shape, skewness):
        """ Markov blanket for the Poisson distribution

        Parameters
        ----------
        y : np.ndarray
            univariate time series

        mean : np.ndarray
            array of location parameters for the Poisson distribution

        scale : float
            scale parameter for the Poisson distribution

        shape : float
            tail thickness parameter for the Poisson distribution

        skewness : float
            skewness parameter for the Poisson distribution

        Returns
        ----------
        - Markov blanket of the Poisson family
        """
        return ss.poisson.logpmf(y, mean)

    @staticmethod
    def setup():
        """ Returns the attributes of this family

        Notes
        ----------
        - scale notes whether family has a variance parameter (sigma)
        - shape notes whether family has a tail thickness parameter (nu)
        - skewness notes whether family has a skewness parameter (gamma)
        - mean_transform is a function which transforms the location parameter
        - cythonized notes whether the family has cythonized routines
        
        Returns
        ----------
        - model name, link function, scale, shape, skewness, mean_transform, cythonized
        """
        name = "Poisson"
        link = np.exp
        scale = False
        shape = False
        skewness = False
        mean_transform = np.log
        cythonized = True
        return name, link, scale, shape, skewness, mean_transform, cythonized

    @staticmethod
    def neg_loglikelihood(y, mean, scale, shape, skewness):
        """ Negative loglikelihood function

        Parameters
        ----------
        y : np.ndarray
            univariate time series

        mean : np.ndarray
            array of location parameters for the Poisson distribution

        scale : float
            scale parameter for the Poisson distribution

        shape : float
            tail thickness parameter for the Poisson distribution

        skewness : float
            skewness parameter for the Poisson distribution

        Returns
        ----------
        - Negative loglikelihood of the Poisson family
        """
        return -np.sum(-mean + np.log(mean)*y - sp.gammaln(y + 1))

    def pdf(self, mu):
        """
        PDF for Poisson prior

        Parameters
        ----------
        mu : float
            Latent variable for which the prior is being formed over

        Returns
        ----------
        - p(mu)
        """
        if self.transform is not None:
            mu = self.transform(mu)                
        return ss.poisson.pmf(mu, self.lmd0) 

    @staticmethod
    def reg_score_function(X, y, mean, scale, shape, skewness):
        """ GAS Poisson Regression Update term using gradient only - native Python function

        Parameters
        ----------
        X : float
            datapoint for the right hand side variable
    
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Poisson distribution

        scale : float
            scale parameter for the Poisson distribution

        shape : float
            tail thickness parameter for the Poisson distribution

        skewness : float
            skewness parameter for the Poisson distribution

        Returns
        ----------
        - Score of the Poisson family
        """
        return X*(y-mean)

    @staticmethod
    def second_order_score(y, mean, scale, shape, skewness):
        """ GAS Poisson Update term potentially using second-order information - native Python function

        Parameters
        ----------
        y : float
            datapoint for the time series

        mean : float
            location parameter for the Poisson distribution

        scale : float
            scale parameter for the Poisson distribution

        shape : float
            tail thickness parameter for the Poisson distribution

        skewness : float
            skewness parameter for the Poisson distribution

        Returns
        ----------
        - Adjusted score of the Poisson family
        """
        return (y-mean)/float(mean)

    # Optional Cythonized recursions below for GAS Poisson models

    @staticmethod
    def gradient_recursion():
        """ GAS Poisson Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Poisson model - gradient only
        """
        return gas_recursion_poisson_orderone

    @staticmethod
    def newton_recursion():
        """ GAS Poisson Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Poisson model - adjusted score
        """
        return gas_recursion_poisson_ordertwo

    @staticmethod
    def gradientx_recursion():
        """ GASX Poisson Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GASX Poisson model - gradient only
        """
        return gasx_recursion_poisson_orderone

    @staticmethod
    def newtonx_recursion():
        """ GASX Poisson Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GASX Poisson model - adjusted score
        """
        return gasx_recursion_poisson_ordertwo

    @staticmethod
    def gradientllev_recursion():
        """ GAS Local Level Poisson Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Local Level Poisson model - gradient only
        """
        return gas_llev_recursion_poisson_orderone

    @staticmethod
    def newtonllev_recursion():
        """ GAS Local Level Poisson Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Local Level Poisson model - adjusted score
        """
        return gas_llev_recursion_poisson_ordertwo

    @staticmethod
    def gradientllt_recursion():
        """ GAS Local Linear Trend Poisson Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Local Linear Trend Poisson model - gradient only
        """
        return gas_llt_recursion_poisson_orderone

    @staticmethod
    def newtonllt_recursion():
        """ GAS Local Linear Trend Poisson Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Local Linear Trend Poisson model - adjusted score
        """
        return gas_llt_recursion_poisson_ordertwo

    @staticmethod
    def gradientreg_recursion():
        """ GAS Dynamic Regression Poisson Model Recursion - gradient only

        Returns
        ----------
        - Recursion function for GAS Dynamic Regression Poisson model - gradient only
        """
        return gas_reg_recursion_poisson_orderone

    @staticmethod
    def newtonreg_recursion():
        """ GAS Dynamic Regression Poisson Model Recursion - adjusted score

        Returns
        ----------
        - Recursion function for GAS Dynamic Regression Poisson model - adjusted score
        """
        return gas_reg_recursion_poisson_ordertwo


import copy
import sys
if sys.version_info < (3,):
    range = xrange

import numpy as np
import pandas as pd
import scipy.stats as ss
from patsy import dmatrices, dmatrix, demo_data

from .. import families as fam
from .. import tsm as tsm
from .. import data_check as dc

from .kalman import *

class DAR(tsm.TSM):
    """ Inherits time series methods from TSM class.

    **** DYNAMIC AUTOREGRESSIVE MODEL ****

    Parameters
    ----------

    ar : int
        Number of autoregressive lags

    data : pd.DataFrame
        Field to specify the data that will be used
    """

    def __init__(self, data, ar, integ=0, target=None):

        # Initialize TSM object
        super(DAR, self).__init__('DAR')

        # Latent Variable information
        self.ar = ar
        self.integ = integ
        self.target = target
        self.model_name = "DAR(" + str(self.ar) + ", integrated=" + str(self.integ) + ")"
        self.max_lag = self.ar
        self._z_hide = 0 # Whether to cutoff latent variables from results table
        self.supported_methods = ["MLE", "PML", "Laplace", "M-H", "BBVI"]
        self.default_method = "MLE"
        self.multivariate_model = False

        # Format the data
        self.data_original = data.copy()
        self.data, self.data_name, self.is_pandas, self.index = dc.data_check(data,target)
        self.data = self.data.astype(np.float) # treat as float for Cython
        self.data_original_nondf = self.data.copy()

        # Difference data
        for order in range(0, self.integ):
            self.data = np.diff(self.data)
            self.data_name = "Differenced " + self.data_name

        self.X = self._ar_matrix()
        self.data = self.data[self.max_lag:]
        self.y = self.data
        self.y_name = self.data_name
        self._create_latent_variables()
        self.z_no = len(self.latent_variables.z_list)

    def _ar_matrix(self):
        """ Creates Autoregressive matrix

        Returns
        ----------
        X : np.ndarray
            Autoregressive Matrix

        """
        Y = np.array(self.data[self.max_lag:self.data.shape[0]])
        X = np.ones(Y.shape[0])

        if self.ar != 0:
            for i in range(0, self.ar):
                X = np.vstack((X,self.data[(self.max_lag-i-1):-i-1]))

        return X.T

    def _create_latent_variables(self):
        """ Creates model latent variables

        Returns
        ----------
        None (changes model attributes)
        """

        self.latent_variables.add_z('Sigma^2 irregular', fam.Flat(transform='exp'), fam.Normal(0,3))

        self.latent_variables.add_z('Constant', fam.Flat(transform=None), fam.Normal(0,3))

        for parm in range(1,self.ar+1):
            self.latent_variables.add_z('Sigma^2 AR(' + str(parm) + ')', fam.Flat(transform='exp'), fam.Normal(0,3))

    def _forecast_model(self,beta,Z,h):
        """ Creates forecasted states and variances

        Parameters
        ----------
        beta : np.ndarray
            Contains untransformed starting values for latent variables

        Returns
        ----------
        a : np.ndarray
            Forecasted states

        P : np.ndarray
            Variance of forecasted states
        """     

        T, _, R, Q, H = self._ss_matrices(beta)
        return dl_univariate_kalman_fcst(self.data,Z,H,T,Q,R,0.0,h)

    def _model(self,data,beta):
        """ Creates the structure of the model

        Parameters
        ----------
        data : np.array
            Contains the time series

        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        a,P,K,F,v : np.array
            Filted states, filtered variances, Kalman gains, F matrix, residuals
        """     

        T, Z, R, Q, H = self._ss_matrices(beta)

        return dl_univariate_kalman(data,Z,H,T,Q,R,0.0)

    def _ss_matrices(self,beta):
        """ Creates the state space matrices required

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        T, Z, R, Q, H : np.array
            State space matrices used in KFS algorithm
        """     

        T = np.identity(self.z_no-1)
        H = np.identity(1)*self.latent_variables.z_list[0].prior.transform(beta[0])       
        Z = self.X
        R = np.identity(self.z_no-1)
        
        Q = np.identity(self.z_no-1)
        for i in range(0,self.z_no-1):
            Q[i][i] = self.latent_variables.z_list[i+1].prior.transform(beta[i+1])

        return T, Z, R, Q, H

    def neg_loglik(self,beta):
        """ Creates the negative log marginal likelihood of the model

        Parameters
        ----------
        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        The negative log logliklihood of the model
        """         
        _, _, _, F, v = self._model(self.y,beta)
        loglik = 0.0
        for i in range(0,self.y.shape[0]):
            loglik += np.linalg.slogdet(F[:,:,i])[1] + np.dot(v[i],np.dot(np.linalg.pinv(F[:,:,i]),v[i]))
        return -(-((self.y.shape[0]/2)*np.log(2*np.pi))-0.5*loglik.T[0].sum())

    def plot_predict(self, h=5, past_values=20, intervals=True, **kwargs):        
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        past_values : int (default : 20)
            How many past observations to show on the forecast graph?

        intervals : Boolean
            Would you like to show 95% prediction intervals for the forecast?

        Returns
        ----------
        - Plot of the forecast
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            y_holder = self.y.copy() # holds past data and predicted data to create AR matrix
            full_X = self.X.copy()
            full_X = np.append(full_X,np.array([np.append(1.0, y_holder[-self.ar:][::-1])]), axis=0)
            Z = full_X

            # Construct Z matrix
            for step in range(h):
                a, P = self._forecast_model(self.latent_variables.get_z_values(),Z,step)
                new_value = np.dot(Z[-1,:],a[:,self.y.shape[0]+step])
                y_holder = np.append(y_holder, new_value)
                Z = np.append(Z, np.array([np.append(1.0, y_holder[-self.ar:][::-1])]), axis=0)

            # Retrieve data, dates and (transformed) latent variables         
            a, P = self._forecast_model(self.latent_variables.get_z_values(),Z,h)
            smoothed_series = np.zeros(self.y.shape[0]+h)
            series_variance = np.zeros(self.y.shape[0]+h)
            for t in range(self.y.shape[0]+h):
                smoothed_series[t] = np.dot(Z[t],a[:,t])
                series_variance[t] = np.dot(np.dot(Z[t],P[:,:,t]),Z[t].T) + self.latent_variables.z_list[0].prior.transform(self.latent_variables.get_z_values()[0])    

            date_index = self.shift_dates(h)
            plot_values = smoothed_series[-h-past_values:]
            forecasted_values = smoothed_series[-h:]
            lower = forecasted_values - 1.98*np.power(series_variance[-h:],0.5)
            upper = forecasted_values + 1.98*np.power(series_variance[-h:],0.5)
            lower = np.append(plot_values[-h-1],lower)
            upper = np.append(plot_values[-h-1],upper)

            plot_index = date_index[-h-past_values:]

            plt.figure(figsize=figsize)
            if intervals == True:
                plt.fill_between(date_index[-h-1:], lower, upper, alpha=0.2)            

            plt.plot(plot_index,plot_values)
            plt.title("Forecast for " + self.y_name)
            plt.xlabel("Time")
            plt.ylabel(self.y_name)
            plt.show()

    def plot_fit(self,intervals=False,**kwargs):
        """ Plots the fit of the model

        Parameters
        ----------
        intervals : Boolean
            Whether to plot 95% confidence interval of states

        Returns
        ----------
        None (plots data and the fit)
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))
        series_type = kwargs.get('series_type','Smoothed')

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            date_index = copy.deepcopy(self.index)
            date_index = date_index[self.integ+self.ar:]

            if series_type == 'Smoothed':
                mu, V = self.smoothed_state(self.data,self.latent_variables.get_z_values())
            elif series_type == 'Filtered':
                mu, V, _, _, _ = self._model(self.data,self.latent_variables.get_z_values())
            else:
                mu, V = self.smoothed_state(self.data,self.latent_variables.get_z_values())

            # Create smoothed/filtered aggregate series
            _, Z, _, _, _ = self._ss_matrices(self.latent_variables.get_z_values())
            smoothed_series = np.zeros(self.y.shape[0])

            for t in range(0,self.y.shape[0]):
                smoothed_series[t] = np.dot(Z[t],mu[:,t])

            plt.figure(figsize=figsize) 
            
            plt.subplot(self.z_no+1, 1, 1)
            plt.title(self.y_name + " Raw and " + series_type)  

            plt.plot(date_index,self.data,label='Data')
            plt.plot(date_index,smoothed_series,label=series_type,c='black')
            plt.legend(loc=2)

            for coef in range(0,self.z_no-1):
                V_coef = V[0][coef][:-1]    
                plt.subplot(self.z_no+1, 1, 2+coef)
                plt.title("Beta " + self.latent_variables.z_list[1+coef].name) 

                if intervals == True:
                    alpha =[0.15*i/float(100) for i in range(50,12,-2)]
                    plt.fill_between(date_index[5:], mu[coef,0:mu.shape[1]-1][5:] + 1.98*np.sqrt(V_coef[5:]), mu[coef,0:mu.shape[1]-1][5:] - 1.98*np.sqrt(V_coef[5:]), alpha=0.15,label='95% C.I.') 
                plt.plot(date_index,mu[coef,0:mu.shape[1]-1],label='Data')
                plt.legend(loc=2)               
            
            plt.subplot(self.z_no+1, 1, self.z_no+1)
            plt.title("Measurement Error")
            plt.plot(date_index,self.data-smoothed_series,label='Irregular')
            plt.legend(loc=2)   

            plt.show()  

    def predict(self, h=5):        
        """ Makes forecast with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps ahead would you like to forecast?

        Returns
        ----------
        - pd.DataFrame with predictions
        """     

        if self.latent_variables.estimated is False:
            raise Exception("No latent variables estimated!")
        else:
            y_holder = self.y.copy() # holds past data and predicted data to create AR matrix
            full_X = self.X.copy()
            full_X = np.append(full_X,np.array([np.append(1.0, y_holder[-self.ar:][::-1])]), axis=0)
            Z = full_X

            for step in range(h):
                a, P = self._forecast_model(self.latent_variables.get_z_values(),Z,step)
                new_value = np.dot(Z[-1,:],a[:,self.y.shape[0]+step])
                y_holder = np.append(y_holder, new_value)
                Z = np.append(Z, np.array([np.append(1.0, y_holder[-self.ar:][::-1])]), axis=0)

            date_index = self.shift_dates(h)

            result = pd.DataFrame(y_holder[-h:])
            result.rename(columns={0:self.y_name}, inplace=True)
            result.index = date_index[-h:]

            return result

    def predict_is(self, h=5, fit_once=True):
        """ Makes dynamic in-sample predictions with the estimated model

        Parameters
        ----------
        h : int (default : 5)
            How many steps would you like to forecast?

        fit_once : boolean
            (default: True) Fits only once before the in-sample prediction; if False, fits after every new datapoint

        Returns
        ----------
        - pd.DataFrame with predicted values
        """     

        predictions = []

        for t in range(0,h):
            data1 = self.data_original_nondf[:-h+t]
            x = DAR(data=data1, ar=self.ar, integ=self.integ)
            if fit_once is False:
                x.fit(printer=False)
            if t == 0:
                if fit_once is True:
                    x.fit(printer=False)
                    saved_lvs = x.latent_variables
                predictions = x.predict(1)
            else:
                if fit_once is True:
                    x.latent_variables = saved_lvs
                predictions = pd.concat([predictions,x.predict(1)])

        predictions.rename(columns={0:self.y_name}, inplace=True)
        predictions.index = self.index[-h:]

        return predictions

    def plot_predict_is(self, h=5, **kwargs):
        """ Plots forecasts with the estimated model against data
            (Simulated prediction with data)

        Parameters
        ----------
        h : int (default : 5)
            How many steps to forecast

        Returns
        ----------
        - Plot of the forecast against data 
        """
        import matplotlib.pyplot as plt
        import seaborn as sns

        figsize = kwargs.get('figsize',(10,7))

        plt.figure(figsize=figsize)
        predictions = self.predict_is(h)
        data = self.data[-h:]
        plt.plot(predictions.index,data,label='Data')
        plt.plot(predictions.index,predictions,label='Predictions',c='black')
        plt.title(self.y_name)
        plt.legend(loc=2)   
        plt.show()          

    def simulation_smoother(self,beta):
        """ Koopman's simulation smoother - simulates from states given
        model parameters and observations

        Parameters
        ----------

        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        - A simulated state evolution
        """         

        T, Z, R, Q, H = self._ss_matrices(beta)

        # Generate e_t+ and n_t+
        rnd_h = np.random.normal(0,np.sqrt(H),self.data.shape[0]+1)
        q_dist = ss.multivariate_normal([0.0, 0.0], Q)
        rnd_q = q_dist.rvs(self.data.shape[0]+1)

        # Generate a_t+ and y_t+
        a_plus = np.zeros((T.shape[0],self.data.shape[0]+1)) 
        a_plus[0,0] = np.mean(self.data[0:5])
        y_plus = np.zeros(self.data.shape[0])

        for t in range(0,self.data.shape[0]+1):
            if t == 0:
                a_plus[:,t] = np.dot(T,a_plus[:,t]) + rnd_q[t,:]
                y_plus[t] = np.dot(Z,a_plus[:,t]) + rnd_h[t]
            else:
                if t != self.data.shape[0]:
                    a_plus[:,t] = np.dot(T,a_plus[:,t-1]) + rnd_q[t,:]
                    y_plus[t] = np.dot(Z,a_plus[:,t]) + rnd_h[t]

        alpha_hat, _ = self.smoothed_state(self.data,beta)
        alpha_hat_plus, _ = self.smoothed_state(y_plus,beta)
        alpha_tilde = alpha_hat - alpha_hat_plus + a_plus
    
        return alpha_tilde

    def smoothed_state(self,data,beta):
        """ Creates the negative log marginal likelihood of the model

        Parameters
        ----------

        data : np.array
            Data to be smoothed

        beta : np.array
            Contains untransformed starting values for latent variables

        Returns
        ----------
        - Smoothed states
        """         

        T, Z, R, Q, H = self._ss_matrices(beta)
        alpha, V = dl_univariate_KFS(data,Z,H,T,Q,R,0.0)
        return alpha, V




import numpy as np
import sys
if sys.version_info < (3,):
    range = xrange

from scipy.stats import multivariate_normal

from .metropolis_sampler import metropolis_sampler

class MetropolisHastings(object):
    """ RANDOM-WALK METROPOLIS-HASTINGS MCMC

    Parameters
    ----------
    posterior : function
        A posterior function

    scale : float
        The scale for the random walk

    nsims : int
        The number of iterations to perform

    initials : np.array
        Where to start the MCMC chain

    cov_matrix : np.array
        (optional) A covariance matrix for the random walk
    
    thinning : int
        By how much to thin the chains (2 means drop every other point)

    warm_up_period : boolean
        Whether to discard first half of the chain as 'warm-up'

    model_object : TSM object
        A model object (for use in SPDK sampling)
    """

    def __init__(self, posterior, scale, nsims, initials, 
        cov_matrix=None, thinning=2, warm_up_period=True, model_object=None):
        self.posterior = posterior
        self.scale = scale
        self.nsims = (1+warm_up_period)*nsims*thinning
        self.initials = initials
        self.param_no = self.initials.shape[0]
        self.phi = np.zeros([self.nsims,self.param_no])
        self.phi[0] = self.initials # point from which to start the Metropolis-Hasting algorithm
        
        if cov_matrix is None:
            self.cov_matrix = np.identity(self.param_no) * np.abs(self.initials)
        else:
            self.cov_matrix = cov_matrix

        self.thinning = thinning
        self.warm_up_period = warm_up_period

        if model_object is not None:
            self.model = model_object

    @staticmethod
    def tune_scale(acceptance, scale):
        """ Tunes scale for M-H algorithm

        Parameters
        ----------
        acceptance : float
            The most recent acceptance rate

        scale : float
            The current scale parameter

        Returns
        ----------
        scale : float
            An adjusted scale parameter

        Notes
        ----------
        Ross : Initially did this by trial and error, then refined by looking at other
        implementations, so some credit here to PyMC3 which became a guideline for this.
        """     

        if acceptance > 0.8:
            scale *= 2.0
        elif acceptance <= 0.8 and acceptance > 0.4:
            scale *= 1.3            
        elif acceptance < 0.234 and acceptance > 0.1:
            scale *= (1/1.3)
        elif acceptance <= 0.1 and acceptance > 0.05:
            scale *= 0.4
        elif acceptance <= 0.05 and acceptance > 0.01:
            scale *= 0.2
        elif acceptance <= 0.01:
            scale *= 0.1
        return scale        

    def sample(self):
        """ Sample from M-H algorithm

        Returns
        ----------
        chain : np.array
            Chains for each parameter

        mean_est : np.array
            Mean values for each parameter

        median_est : np.array
            Median values for each parameter

        upper_95_est : np.array
            Upper 95% credibility interval for each parameter

        lower_95_est : np.array
            Lower 95% credibility interval for each parameter           
        """     

        acceptance = 1
        finish = 0

        while (acceptance < 0.234 or acceptance > 0.4) or finish == 0:

            # If acceptance is in range, proceed to sample, else continue tuning
            if not (acceptance < 0.234 or acceptance > 0.4):
                finish = 1
                print("")
                print("Tuning complete! Now sampling.")
                sims_to_do = self.nsims
            else:
                sims_to_do = int(self.nsims/2) # For acceptance rate tuning

            # Holds data on acceptance rates and uniform random numbers
            a_rate = np.zeros([sims_to_do,1])
            crit = np.random.rand(sims_to_do,1)
            post = multivariate_normal(np.zeros(self.param_no), self.cov_matrix)
            rnums = post.rvs()*self.scale
            
            for k in range(1,sims_to_do): 
                rnums = np.vstack((rnums,post.rvs()*self.scale))

            self.phi, a_rate = metropolis_sampler(sims_to_do, self.phi, self.posterior, 
                a_rate, rnums, crit)

            acceptance = a_rate.sum()/a_rate.shape[0]
            self.scale = self.tune_scale(acceptance,self.scale)

            print("Acceptance rate of Metropolis-Hastings is " + str(acceptance))

        # Remove warm-up and thin
        self.phi = self.phi[self.nsims/2:,:][::self.thinning,:]

        chain = np.array([self.phi[i][0] for i in range(0, self.phi.shape[0])])

        for m in range(1, self.param_no):
            chain = np.vstack((chain, [self.phi[i][m] for i in range(0,self.phi.shape[0])]))

        if self.param_no == 1:
            chain = np.array([chain])

        mean_est = np.array([np.mean(np.array([self.phi[i][j] for i in range(0,self.phi.shape[0])])) for j in range(self.param_no)])
        median_est = np.array([np.median(np.array([self.phi[i][j] for i in range(0,self.phi.shape[0])])) for j in range(self.param_no)])
        upper_95_est = np.array([np.percentile(np.array([self.phi[i][j] for i in range(0,self.phi.shape[0])]), 95) for j in range(self.param_no)])
        lower_95_est = np.array([np.percentile(np.array([self.phi[i][j] for i in range(0,self.phi.shape[0])]), 5) for j in range(self.param_no)])

        return chain, mean_est, median_est, upper_95_est, lower_95_est